#!/usr/bin/env python
"""
AWS Level 0 (L0) to Level 1 (L1) data processing
"""
import numpy as np
import pandas as pd
import xarray as xr
import re, logging
from pypromice.process.value_clipping import clip_values
logger = logging.getLogger(__name__)


def toL1(L0, vars_df, T_0=273.15, tilt_threshold=-100):
    '''Process one Level 0 (L0) product to Level 1

    Parameters
    ----------
    L0 : xarray.Dataset
        Level 0 dataset
    vars_df : pd.DataFrame
        Metadata dataframe
    T_0 : int
        Air temperature for sonic ranger adjustment
    tilt_threshold : int
        Tilt-o-meter threshold for valid measurements
        
    Returns
    -------
    ds : xarray.Dataset
        Level 1 dataset
    '''    
    assert(type(L0) == xr.Dataset)
    ds = L0
    ds.attrs['level'] = 'L1'

    for l in list(ds.keys()):
        if l not in ['time', 'msg_i', 'gps_lat', 'gps_lon', 'gps_alt', 'gps_time']:
            ds[l] = _reformatArray(ds[l])

    # ds['time_orig'] = ds['time'] # Not used

    # The following drops duplicate datetime indices. Needs to run before _addTimeShift!
    # We can optionally also drop duplicates within _addTimeShift using pandas duplicated,
    # but retaining the following code instead to preserve previous methods. PJW
    _, index = np.unique(ds['time'], return_index=True)
    ds = ds.isel(time=index)

    # If we do not want to shift hourly average values back -1 hr, then comment the following line.
    ds = addTimeShift(ds, vars_df)

    if hasattr(ds, 'dsr_eng_coef'): 
        ds['dsr'] = (ds['dsr'] * 10) / ds.attrs['dsr_eng_coef']                # Convert radiation from engineering to physical units
    if hasattr(ds, 'usr_eng_coef'):                                            # TODO add metadata to indicate whether radiometer values are corrected with calibration values or not
        ds['usr'] = (ds['usr'] * 10) / ds.attrs['usr_eng_coef']
    if hasattr(ds, 'dlr_eng_coef'):
        ds['dlr'] = ((ds['dlr'] * 10) / ds.attrs['dlr_eng_coef']) + 5.67E-8*(ds['t_rad'] + T_0)**4
    if hasattr(ds, 'ulr_eng_coef'):
        ds['ulr'] = ((ds['ulr'] * 10) / ds.attrs['ulr_eng_coef']) + 5.67E-8*(ds['t_rad'] + T_0)**4

    ds['z_boom_u'] = _reformatArray(ds['z_boom_u'])                            # Reformat boom height
    
    ds['t_u_interp'] = interpTemp(ds['t_u'], vars_df)
    ds['z_boom_u'] = ds['z_boom_u'] * ((ds['t_u_interp'] + T_0)/T_0)**0.5      # Adjust sonic ranger readings for sensitivity to air temperature       
    
    if ds['gps_lat'].dtype.kind == 'O':                                        # Decode and reformat GPS information
        if 'NH' in ds['gps_lat'].dropna(dim='time').values[1]:
            ds = decodeGPS(ds, ['gps_lat','gps_lon','gps_time'])
        elif 'L' in ds['gps_lat'].dropna(dim='time').values[1]:
            logger.info('Found L in GPS string')
            ds = decodeGPS(ds, ['gps_lat','gps_lon','gps_time'])
            for l in ['gps_lat', 'gps_lon']:
                ds[l] = ds[l]/100000
        else:
            try:
                ds = decodeGPS(ds, ['gps_lat','gps_lon','gps_time'])          # TODO this is a work around specifically for L0 RAW processing for THU_U. Find a way to make this slicker
            
            except:
                print('Invalid GPS type {ds["gps_lat"].dtype} for decoding')
            
    for l in ['gps_lat', 'gps_lon', 'gps_alt','gps_time']:
        ds[l] = _reformatArray(ds[l])  

    if hasattr(ds, 'latitude') and hasattr(ds, 'longitude'):
        ds['gps_lat'] = reformatGPS(ds['gps_lat'], ds.attrs['latitude'])
        ds['gps_lon'] = reformatGPS(ds['gps_lon'], ds.attrs['longitude'])

    if hasattr(ds, 'logger_type'):                                             # Convert tilt voltage to degrees
        if ds.attrs['logger_type'].upper() == 'CR1000':                    
            ds['tilt_x']  = getTiltDegrees(ds['tilt_x'], tilt_threshold) 
            ds['tilt_y'] = getTiltDegrees(ds['tilt_y'], tilt_threshold)  
            
    if hasattr(ds, 'tilt_y_factor'):                                           # Apply tilt factor (e.g. -1 will invert tilt angle)
        ds['tilt_y'] = ds['tilt_y']*ds.attrs['tilt_y_factor']

    # Smooth everything
    # Note that this should be OK for CR1000 tx (data only every 6 hrs),
    # since we interpolate above in _getTiltDegrees. PJW
    ds['tilt_x']  = smoothTilt(ds['tilt_x'], 7)                                # Smooth tilt
    ds['tilt_y']  = smoothTilt(ds['tilt_y'], 7)                               
    
    if hasattr(ds, 'bedrock'):                                                 # Fix tilt to zero if station is on bedrock
        if ds.attrs['bedrock']==True or ds.attrs['bedrock'].lower() in 'true':
            ds.attrs['bedrock'] = True                                         # ensures all AWS objects have a 'bedrock' attribute
            ds['tilt_x'] = (('time'), np.arange(ds['time'].size)*0)
            ds['tilt_y'] = (('time'), np.arange(ds['time'].size)*0)
        else:
            ds.attrs['bedrock'] = False                                        # ensures all AWS objects have a 'bedrock' attribute
    else:
        ds.attrs['bedrock'] = False                                            # ensures all AWS objects have a 'bedrock' attribute
    
    if ds.attrs['number_of_booms']==1:                                         # 1-boom processing
        if ~ds['z_pt'].isnull().all():                                         # Calculate pressure transducer fluid density                                           
            if hasattr(ds, 'pt_z_offset'):                                     # Apply SR50 stake offset
                ds['z_pt'] = ds['z_pt'] + int(ds.attrs['pt_z_offset'])              
            ds['z_pt_cor'],ds['z_pt']=getPressDepth(ds['z_pt'], ds['p_u'], 
                                                    ds.attrs['pt_antifreeze'], 
                                                    ds.attrs['pt_z_factor'], 
                                                    ds.attrs['pt_z_coef'], 
                                                    ds.attrs['pt_z_p_coef'])       
        ds['z_stake'] = _reformatArray(ds['z_stake'])                          # Reformat boom height    
        ds['z_stake'] = ds['z_stake'] * ((ds['t_u'] + T_0)/T_0)**0.5           # Adjust sonic ranger readings for sensitivity to air temperature
        
    elif ds.attrs['number_of_booms']==2:                                       # 2-boom processing
        ds['z_boom_l'] = _reformatArray(ds['z_boom_l'])                        # Reformat boom height    
        ds['t_l_interp'] = interpTemp(ds['t_l'], vars_df)
        ds['z_boom_l'] = ds['z_boom_l'] * ((ds['t_l_interp']+ T_0)/T_0)**0.5   # Adjust sonic ranger readings for sensitivity to air temperature    

    ds = clip_values(ds, vars_df)
    for key in ['hygroclip_t_offset', 'dsr_eng_coef', 'usr_eng_coef',
          'dlr_eng_coef', 'ulr_eng_coef', 'pt_z_coef', 'pt_z_p_coef',
          'pt_z_factor', 'pt_antifreeze', 'boom_azimuth', 'nodata',
          'conf', 'file']:
        ds.attrs.pop(key, None)

    return ds

def addTimeShift(ds, vars_df):
    '''Shift times based on file format and logger type (shifting only hourly averaged values,
    and not instantaneous variables). For raw (10 min), all values are sampled instantaneously
    so do not shift. For STM (1 hour), values are averaged and assigned to end-of-hour by the
    logger, so shift by -1 hr. For TX (time frequency depends on v2 or v3) then time is shifted
    depending on logger type. We use the 'instantaneous_hourly' boolean from variables.csv to
    determine if a variable is considered instantaneous at hourly samples.

    This approach creates two separate sub-dataframes, one for hourly-averaged variables
    and another for instantaneous variables. The instantaneous dataframe should never be
    shifted. We apply shifting only to the hourly average dataframe, then concat the two
    dataframes back together.

    It is possible to use pandas merge or join instead of concat, there are equivalent methods
    in each. In this case, we use concat throughout.

    Fausto et al. 2021 specifies the convention of assigning hourly averages to start-of-hour,
    so we need to retain this unless clearly communicated to users.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to apply time shift to
    vars_df : pd.DataFrame
        Metadata dataframe

    Returns
    -------
    ds_out : xarray.Dataset
        Dataset with shifted times
    '''
    df = ds.to_dataframe()
    # No need to drop duplicates here if performed prior to calling this function.
    # df = df[~df.index.duplicated(keep='first')] # drop duplicates, keep=first is arbitrary
    df['doy'] = df.index.dayofyear
    i_cols = [x for x in df.columns if x in vars_df.index and vars_df['instantaneous_hourly'][x] is True] # instantaneous only, list of columns
    df_i = df.filter(items=i_cols, axis=1) # instantaneous only dataframe
    df_a = df.drop(df_i.columns, axis=1) # hourly ave dataframe

    if ds.attrs['format'] == 'raw':
        # 10-minute data, no shifting
        df_out = df
    elif ds.attrs['format'] == 'STM':
        # hourly-averaged, non-transmitted
        # shift everything except instantaneous, any logger type
        df_a = df_a.shift(periods=-1, freq="h")
        df_out = pd.concat([df_a, df_i], axis=1) # different columns, same datetime indices
        df_out = df_out.sort_index()
    elif ds.attrs['format'] == 'TX':
        if ds.attrs['logger_type'] == 'CR1000X':
            # v3, data is hourly all year long
            # shift everything except instantaneous
            df_a = df_a.shift(periods=-1, freq="h")
            df_out = pd.concat([df_a, df_i], axis=1) # different columns, same datetime indices
            df_out = df_out.sort_index()
        elif ds.attrs['logger_type'] == 'CR1000':
            # v2, data is hourly (6-hr for instantaneous) for DOY 100-300, otherwise daily at 00 UTC
            # shift non-instantaneous hourly for DOY 100-300, else do not shift daily
            df_a_hourly = df_a.loc[(df_a['doy'] >= 100) & (df_a['doy'] <= 300)]
            # df_a_hourly = df_a.loc[df_a['doy'].between(100, 300, inclusive='both')] # equivalent to above
            df_a_daily_1 = df_a.loc[(df_a['doy'] < 100)]
            df_a_daily_2 = df_a.loc[(df_a['doy'] > 300)]

            # shift the hourly ave data
            df_a_hourly = df_a_hourly.shift(periods=-1, freq="h")

            # stitch everything back together
            df_concat_u = pd.concat([df_a_daily_1, df_a_daily_2, df_a_hourly], axis=0) # same columns, different datetime indices
            # It's now possible for df_concat_u to have duplicate datetime indices
            df_concat_u = df_concat_u[~df_concat_u.index.duplicated(keep='first')] # drop duplicates, keep=first is arbitrary

            df_out = pd.concat([df_concat_u, df_i], axis=1) # different columns, same datetime indices
            df_out = df_out.sort_index()

    # Back to xarray, and re-assign the original attrs
    df_out = df_out.drop('doy', axis=1)
    ds_out = df_out.to_xarray()
    ds_out = ds_out.assign_attrs(ds.attrs) # Dataset attrs
    for x in ds_out.data_vars: # variable-specific attrs
        ds_out[x].attrs = ds[x].attrs

    # equivalent to above:
    # vals = [xr.DataArray(data=df_out[c], dims=['time'], coords={'time':df_out.index}, attrs=ds[c].attrs) for c in df_out.columns]
    # ds_out = xr.Dataset(dict(zip(df_out.columns, vals)), attrs=ds.attrs)
    return ds_out

def getPressDepth(z_pt, p, pt_antifreeze, pt_z_factor, pt_z_coef, pt_z_p_coef): 
    '''Adjust pressure depth and calculate pressure transducer depth based on 
    pressure transducer fluid density
    
    Parameters
    ----------
    z_pt : xr.Dataarray
        Pressure transducer height (corrected for offset)
    p : xr.Dataarray
        Air pressure
    pt_antifreeze : float
        Pressure transducer anti-freeze percentage for fluid density 
        correction
    pt_z_factor : float
        Pressure transducer factor
    pt_z_coef : float
        Pressure transducer coefficient
    pt_z_p_coef : float
        Pressure transducer coefficient
    
    Returns
    -------
    z_pt_cor : xr.Dataarray
        Pressure transducer height corrected
    z_pt : xr.Dataarray
        Pressure transducer depth
    '''
    # Calculate pressure transducer fluid density                                        
    if pt_antifreeze == 50:                                                    #TODO: Implement function w/ reference (analytical or from LUT)                                             
        rho_af = 1092                                                          #TODO: Track uncertainty
    elif pt_antifreeze == 100:
        rho_af = 1145
    else:
        rho_af = np.nan
        logger.info('ERROR: Incorrect metadata: "pt_antifreeze" = ' +
              f'{pt_antifreeze}. Antifreeze mix only supported at 50% or 100%')
        # assert(False)
                
    # Correct pressure depth
    z_pt_cor = z_pt * pt_z_coef * pt_z_factor * 998.0 / rho_af + 100 * (pt_z_p_coef - p) / (rho_af * 9.81)

    # Calculate pressure transducer depth
    z_pt = z_pt * pt_z_coef * pt_z_factor * 998.0 / rho_af
    
    return z_pt_cor, z_pt


def interpTemp(temp, var_configurations, max_interp=pd.Timedelta(12,'h')):
    '''Clip and interpolate temperature dataset for use in corrections
    
    Parameters
    ----------
    temp : `xarray.DataArray`
        Array of temperature data
    vars_df : `pandas.DataFrame`
        Dataframe to retrieve attribute hi-lo values from for temperature clipping
    max_interp : `pandas.Timedelta`
        Maximum time steps to interpolate across. The default is 12 hours.
    
    Returns
    -------
    temp_interp : `xarray.DataArray`
        Array of interpolatedtemperature data
    '''
    # Determine if upper or lower temperature array
    var = temp.name.lower()
    
    # Find range threshold and use it to clip measurements
    cols = ["lo", "hi", "OOL"]
    assert set(cols) <= set(var_configurations.columns)
    variable_limits = var_configurations[cols].dropna(how="all")
    temp = temp.where(temp >= variable_limits.loc[var,'lo'])
    temp = temp.where(temp <= variable_limits.loc[var, 'hi'])
    
    # Drop duplicates and interpolate across NaN values
#    temp_interp = temp.drop_duplicates(dim='time', keep='first')
    temp_interp = temp.interpolate_na(dim='time', max_gap=max_interp)
    
    return temp_interp


def smoothTilt(tilt, win_size):
    '''Smooth tilt values using a rolling window. This is translated from the
    previous IDL/GDL smoothing algorithm:
    tiltX = smooth(tiltX,7,/EDGE_MIRROR,MISSING=-999) & tiltY = smooth(tiltY,7,/EDGE_MIRROR, MISSING=-999)
    endif
    In Python, this should be
    dstxy = dstxy.rolling(time=7, win_type='boxcar', center=True).mean()
    But the EDGE_MIRROR makes it a bit more complicated
    
    Parameters
    ----------
    tilt : xarray.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (can be in degrees or voltage)
    win_size : int
        Window size to use in pandas 'rolling' method.
        e.g. a value of 7 spans 70 minutes using 10 minute data.

    Returns
    -------
    tdf_rolling : tuple, as: (str, numpy.ndarray)
        The numpy array is the tilt values, smoothed with a rolling mean
    '''
    s = int(win_size/2)
    tdf = tilt.to_dataframe()
    mirror_start = tdf.iloc[:s][::-1]
    mirror_end = tdf.iloc[-s:][::-1]
    mirrored_tdf = pd.concat([mirror_start, tdf, mirror_end])

    tdf_rolling = (
        ('time'),
        mirrored_tdf.rolling(
            win_size, win_type='boxcar', min_periods=1, center=True
            ).mean()[s:-s].values.flatten()
        )
    return tdf_rolling

def getTiltDegrees(tilt, threshold):
    '''Filter tilt with given threshold, and convert from voltage to degrees. 
    Voltage-to-degrees converseion is based on the equation in 3.2.9 at 
    https://essd.copernicus.org/articles/13/3819/2021/#section3    

    Parameters
    ----------
    tilt : xarray.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (voltage)
    threshold : int
        Values below this threshold (-100) will not be retained.
    
    Returns
    -------
    dst.interpolate_na() : xarray.DataArray
        Array (either 'tilt_x' or 'tilt_y'), tilt values (degrees)
    '''
    # notOKtiltX = where(tiltX lt -100, complement=OKtiltX) & notOKtiltY = where(tiltY lt -100, complement=OKtiltY)
    notOKtilt = (tilt < threshold)
    OKtilt = (tilt >= threshold)
    tilt = tilt / 10
    
    # IDL version:
    # tiltX = tiltX/10.
    # tiltnonzero = where(tiltX ne 0 and tiltX gt -40 and tiltX lt 40)
    # if n_elements(tiltnonzero) ne 1 then tiltX[tiltnonzero] = tiltX[tiltnonzero]/abs(tiltX[tiltnonzero])*(-0.49*(abs(tiltX[tiltnonzero]))^4 + 3.6*(abs(tiltX[tiltnonzero]))^3 - 10.4*(abs(tiltX[tiltnonzero]))^2 +21.1*(abs(tiltX[tiltnonzero])))
    # tiltY = tiltY/10.
    # tiltnonzero = where(tiltY ne 0 and tiltY gt -40 and tiltY lt 40)
    # if n_elements(tiltnonzero) ne 1 then tiltY[tiltnonzero] = tiltY[tiltnonzero]/abs(tiltY[tiltnonzero])*(-0.49*(abs(tiltY[tiltnonzero]))^4 + 3.6*(abs(tiltY[tiltnonzero]))^3 - 10.4*(abs(tiltY[tiltnonzero]))^2 +21.1*(abs(tiltY[tiltnonzero])))
    
    dst = tilt
    nz = (dst != 0) & (np.abs(dst) < 40)
    
    dst = dst.where(~nz, other = dst / np.abs(dst)
                      * (-0.49
                         * (np.abs(dst))**4 + 3.6
                         * (np.abs(dst))**3 - 10.4
                         * (np.abs(dst))**2 + 21.1
                         * (np.abs(dst))))
    
    # if n_elements(OKtiltX) gt 1 then tiltX[notOKtiltX] = interpol(tiltX[OKtiltX],OKtiltX,notOKtiltX) ; Interpolate over gaps for radiation correction; set to -999 again below.
    dst = dst.where(~notOKtilt)
    return dst.interpolate_na(dim='time', use_coordinate=False)                #TODO: Filling w/o considering time gaps to re-create IDL/GDL outputs. Should fill with coordinate not False. Also consider 'max_gap' option?

    
def decodeGPS(ds, gps_names):
    '''Decode GPS information based on names of GPS attributes. This should be 
    applied if gps information does not consist of float values
    
    Parameters
    ----------
    ds : xr.Dataset
        Data set
    gps_names : list
        Variable names for GPS information, such as "gps_lat", "gps_lon" and
        "gps_alt"
    
    Returns
    -------
    ds : xr.Dataset
        Data set with decoded GPS information
    '''
    for v in gps_names:
        a = ds[v].attrs    
        str2nums = [re.findall(r"[-+]?\d*\.\d+|\d+", _) if isinstance(_, str) else [np.nan] for _ in ds[v].values]
        ds[v][:] = pd.DataFrame(str2nums).astype(float).T.values[0]
        ds[v] = ds[v].astype(float)
        ds[v].attrs = a 
    return ds

def reformatGPS(pos_arr, attrs):
    '''Correct latitude and longitude from native format to decimal degrees.
    
    v2 stations should send  "NH6429.01544","WH04932.86061" (NUK_L 2022)
    v3 stations should send coordinates as "6628.93936","04617.59187" (DY2) or 6430,4916 (NUK_Uv3)
    decodeGPS should have decoded these strings to floats in ddmm.mmmm format
    v1 stations however only saved decimal minutes (mm.mmmmm) as float<=60. '
    In this case, we use the integer part of the latitude given in the config 
    file and append the gps value after it.
        
    Parameters
    ----------
    pos_arr : xr.Dataarray
        Array of latitude or longitude measured by the GPS
    attrs : dict
        The global attribute 'latitude' or 'longitude' associated with the 
        file being processed. It is the standard latitude/longitude given in the 
        config file for that station.
    
    Returns
    -------
    pos_arr : xr.Dataarray
        Formatted GPS position array in decimal degree
    '''       
    if np.any((pos_arr <= 90) & (pos_arr > 0)):  
        # then pos_arr is in decimal minutes, so we add to it the integer 
        # part of the latitude given in the config file x100
        # so that it reads ddmm.mmmmmm like for v2 and v3 files
        # Note that np.sign and np.attrs handles negative longitudes.
        pos_arr = np.sign(attrs) * (pos_arr + 100*np.floor(np.abs(attrs)))
    a = pos_arr.attrs                                                     
    pos_arr = np.floor(pos_arr / 100) + (pos_arr / 100 - np.floor(pos_arr / 100)) * 100 / 60
    pos_arr.attrs = a 
    return pos_arr 

def _reformatArray(ds_arr):
    '''Reformat DataArray values and attributes
    
    Parameters
    ----------
    ds_arr : xr.Dataarray
        Data array
    
    Returns
    -------
    ds_arr : xr.Dataarray
        Formatted data array
    '''
    a = ds_arr.attrs                                                           # Store
    ds_arr.values = pd.to_numeric(ds_arr, errors='coerce')
    ds_arr.attrs = a                                                           # Reformat
    return ds_arr         

def _removeVars(ds, v_names):
    '''Remove redundant variables if present in dataset
    
    Parameters
    ----------
    ds : xr.Dataset
        Data set
    v_names : list
        List of column names to drop
    
    Returns
    -------
    ds : xr.Dataset
        Data set with removed variables
    '''
    for v in v_names:
        if v in list(ds.variables): ds = ds.drop_vars(v)
    return ds

def _popCols(ds, booms, data_type, vars_df, cols):
    '''Populate data array columns with given variable names from look-up table
    
    Parameters
    ----------
    ds : xr.Dataset
        Data set
    booms : int
        Number of booms (1 or 2)
    data_type : str
        Type of data ("tx", "raw")
    vars_df : pd.DataFrame
        Variables lookup table
    cols : list
        Names of columns to populate
    
    Returns
    -------
    ds : xr.Dataset
        Data with populated columns
    '''
    if booms==1:
        names = vars_df.loc[(vars_df[cols[0]]!='two-boom')]

    elif booms==2:
        names = vars_df.loc[(vars_df[cols[0]]!='one-boom')]
       
    for v in list(names.index):
        if v not in list(ds.variables):
            ds[v] = (('time'), np.arange(ds['time'].size)*np.nan)      
    return ds

# def _popCols(ds, booms, data_type, vars_df, cols):
#     if booms==1:
#         if data_type !='TX':
#             names = vars_df.loc[(vars_df[cols[0]]!='two-boom')]
#         else:
#             names = vars_df.loc[(vars_df[cols[0]] != 'two-boom') & vars_df[cols[1]] != 'tx']
    
#     elif booms==2:
#         if data_type !='TX':
#             names = vars_df.loc[(vars_df[cols[0]]!='two-boom')]
#         else:
#             names = vars_df.loc[(vars_df[cols[0]] != 'two-boom') & vars_df[cols[1]] != 'tx']
       
#     for v in list(names.index):
#         if v not in list(ds.variables):
#             ds[v] = (('time'), np.arange(ds['time'].size)*np.nan)      
#     return ds

#------------------------------------------------------------------------------

if __name__ == "__main__": 
    # unittest.main() 
    pass    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Load module
"""
from datetime import timedelta
from typing import Sequence, Optional

import logging
import os
import pandas as pd
import toml
import xarray as xr

logger = logging.getLogger(__name__)


def getConfig(
    config_file, inpath, default_columns: Sequence[str] = ("msg_lat", "msg_lon")
):
    """Load configuration from .toml file. PROMICE .toml files support defining
    features at the top level which apply to all nested properties, but do not
    overwrite nested properties if they are defined

    Parameters
    ----------
    config_file : str
        TOML file path
    inpath : str
        Input folder directory where L0 files can be found

    Returns
    -------
    conf : dict
        Configuration dictionary
    """
    conf = toml.load(config_file)  # Move all top level keys to nested properties,
    top = [
        _ for _ in conf.keys() if not type(conf[_]) is dict
    ]  # if they are not already defined in the nested properties
    subs = [
        _ for _ in conf.keys() if type(conf[_]) is dict
    ]  # Insert the section name (config_file) as a file property and config file
    for s in subs:
        for t in top:
            if t not in conf[s].keys():
                conf[s][t] = conf[t]

        conf[s]["conf"] = config_file
        conf[s]["file"] = os.path.join(inpath, s)
        conf[s]["columns"].extend(default_columns)

    for t in top:
        conf.pop(t)  # Delete all top level keys beause each file
    # should carry all properties with it
    for k in conf.keys():  # Check required fields are present
        for field in ["columns", "station_id", "format", "skiprows"]:
            assert field in conf[k].keys(), field + " not in config keys"
    return conf


def getL0(
    infile: str,
    nodata,
    cols,
    skiprows,
    file_version,
    delimiter=",",
    comment="#",
    time_offset: Optional[float] = None,
) -> xr.Dataset:
    """Read L0 data file into pandas DataFrame object

    Parameters
    ----------
    infile : str
        L0 file path
    nodata : list
        List containing value for nan values and reassigned value
    cols : list
        List of columns in file
    skiprows : int
        Skip rows value
    file_version : int
        Version of L0 file
    delimiter : str
        String delimiter for L0 file
    comment : str
        Notifier of commented sections in L0 file
    time_offset : Optional[float]
        Time offset in hours for correcting for non utc time data.
    Returns
    -------
    ds : xarray.Dataset
        L0 Dataset
    """
    if file_version == 1:
        df = pd.read_csv(
            infile,
            comment=comment,
            index_col=0,
            na_values=nodata,
            names=cols,
            sep=delimiter,
            skiprows=skiprows,
            skip_blank_lines=True,
            usecols=range(len(cols)),
            low_memory=False,
        )
        df["time"] = pd.to_datetime(
            df.year.astype(str)
            + df.doy.astype(str).str.zfill(3)
            + df.hhmm.astype(str).str.zfill(4),
            format="%Y%j%H%M",
        )
        df = df.set_index("time")

    else:
        df = pd.read_csv(
            infile,
            comment=comment,
            index_col=0,
            na_values=nodata,
            names=cols,
            parse_dates=True,
            sep=delimiter,
            skiprows=skiprows,
            skip_blank_lines=True,
            usecols=range(len(cols)),
            low_memory=False,
        )
        try:
            df.index = pd.to_datetime(df.index)
        except ValueError as e:
            logger.info("\n" + infile)
            logger.info("\nValueError:")
            logger.info(e)
            logger.info("\t\t> Trying pd.to_datetime with format=mixed")
            try:
                df.index = pd.to_datetime(df.index, format="mixed")
            except Exception as e:
                logger.info("\nDateParseError:")
                logger.info(e)
                logger.info(
                    "\t\t> Trying again removing apostrophes in timestamp (old files format)"
                )
                df.index = pd.to_datetime(df.index.str.replace('"', ""))

    if time_offset is not None:
        df.index = df.index + timedelta(hours=time_offset)

    # Drop SKIP columns
    for c in df.columns:
        if c[0:4] == "SKIP":
            df.drop(columns=c, inplace=True)

    # Carry relevant metadata with ds
    ds = xr.Dataset.from_dataframe(df)
    ds.attrs["level"] = "L0"

    return ds



import argparse
from pathlib import Path

import pandas as pd

from pypromice.postprocess.bufr_utilities import read_bufr_file


def main():
    parser = argparse.ArgumentParser("BUFR to CSV converter")
    parser.add_argument("path", type=Path, nargs='+')
    args = parser.parse_args()

    paths = []
    for path in args.path:
        paths += list(path.parent.glob(path.name))

    df = pd.concat([read_bufr_file(path) for path in paths])
    print(df.to_csv())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AWS data retrieval module
"""
from pyDataverse.api import NativeApi
import pandas as pd
import xarray as xr
import unittest, pkg_resources
from datetime import datetime
import warnings, os

def aws_names():
    '''Return PROMICE and GC-Net AWS names that can be used in get.aws_data() 
    fetching'''
    lookup = lookup_table(['doi:10.22008/FK2/IW73UU'])
    print(f'Available dataset keywords: {list(lookup.keys())}')
    return list(lookup.keys())

def aws_data(aws_name):
    '''Return PROMICE and GC-Net AWS L3 v3 hourly observations
    
    Returns
    -------
    df : pandas.DataFrame
        AWS observations dataframe
    '''
    lookup = lookup_table(['doi:10.22008/FK2/IW73UU'])
    assert aws_name.lower() in list(lookup.keys())
    data = pd.read_csv(lookup[aws_name], index_col=0, parse_dates=True)
    return data        
  
def watson_discharge(t='hour'):
    '''Return PROMICE hourly Watson river discharge
    
    Parameters
    ----------
    t : str
        Temporal resolution of the data - "hour", "day" or "year"

    Returns
    -------
    df : pandas.DataFrame
        Watson river discharge dataframe    
    '''
    lookup = lookup_table(['doi:10.22008/FK2/XEHYCM'])
    
    dict_keys = lookup.keys()
    
    if 'year' in t.lower():
        
        key = [k for k in dict_keys if 'year' in k]
        
        if not key: 
            warnings.warn('The yearly Watson River Discharge file does not exist, or has changed name, on GEUS Dataverse DOI, ' + \
                  'please check the dataset, and the naming of the txt files on Dataverse')
                
        if len(key) > 1: 
            warnings.warn('Warning, there exist multiple yearly txt files on dataverse, please check ' + \
                  'if the correct txt file is used')
        
        link = lookup[key[0]]
        df = pd.read_csv(link, sep="\s+", skiprows=9, index_col=0)
        
    elif 'daily' in t.lower() or 'day' in t.lower():
        
        key = [k for k in dict_keys if 'daily' in k]
        
        if not key: 
            warnings.warn('The daily Watson River Discharge file does not exist, or has changed name, on GEUS Dataverse DOI, ' + \
                  'please check the dataset, and the naming of the txt files on Dataverse')
                
        if len(key) > 1: 
            warnings.warn('Warning, there exist multiple daily txt files on dataverse, please check ' + \
                  'if the correct txt file is used')
        
        link = lookup[key[0]]   
        
        df = pd.read_csv(link, sep="\s+", parse_dates=[[0,1,2]])\
                        .rename({"WaterFluxDiversOnly(m3/s)"        : "divers",
                                "Uncertainty(m3/s)"                 : "divers_err",
                                "WaterFluxDivers&Temperature(m3/s)" : "divers_t",
                                "Uncertainty(m3/s).1"               : "divers_t_err",
                                "WaterFluxCumulative(km3)"          : "Q",
                                "Uncertainty(km3)"                  : "err"}, 
                                axis='columns')
        df['time'] = df.iloc[:,0]
        df = df.set_index('time')
        df.drop(columns=df.columns[0:1], axis=1, inplace=True) 
        
    else:
        
        key = [k for k in dict_keys if 'hourly' in k]
        
        if not key: 
            warnings.warn('The hourly Watson River Discharge file does not exist, or has changed name, on GEUS Dataverse DOI, ' + \
                  'please check the dataset, and the naming of the txt files on Dataverse')
                
        if len(key) > 1: 
            warnings.warn('Warning, there exist multiple Houlry txt files on dataverse, please check ' + \
                  'if the correct txt file is used')
        
        link = lookup[key[0]]   
        
        df = pd.read_csv(link, sep="\s+", parse_dates=[[0,1,2,3]])\
                        .rename({"WaterFluxDiversOnly(m3/s)"        : "divers",
                                "Uncertainty(m3/s)"                 : "divers_err",
                                "WaterFluxDivers&Temperature(m3/s)" : "divers_t",
                                "Uncertainty(m3/s).1"               : "divers_t_err",
                                "WaterFluxCumulative(km3)"          : "Q",
                                "Uncertainty(km3)"                  : "err"}, 
                                axis='columns')
        df = _getDFdatetime(df, list(df.iloc[:,0]))            

    return df

def lookup_table(base_dois,
                 server='https://dataverse.geus.dk'):
    '''Fetch dictionary of data files and download URLs from a DOI entry in the
    GEUS Dataverse
    
    Parameters
    ----------
    base_dois : list
        List of DOIs to search
    server : str, optional
        DOI server. The default is "https://dataverse.geus.dk"
    '''
    # Prime API
    dataverse_server = server.strip("/")
    api = NativeApi(dataverse_server)
    
    # Look through DOI entries
    lookup_list = {}
    for d in base_dois:
        dataset = api.get_dataset(d)
        
        # Get file names and DOIs
        f_list = dataset.json()['data']['latestVersion']['files']
        for f in f_list:
            fname = f['dataFile']['filename'].lower()
            if '.csv' in fname or '.txt' in fname:
                link = _getURL(f['dataFile']['persistentId'])
                lookup_list[fname] = link
    return lookup_list 

def _getURL(persistentId, 
            base_link='https://dataverse.geus.dk/api/access/datafile/:persistentId?persistentId='):
    '''Return URL download link from persistentId attached to DOI'''
    return base_link+persistentId  


def _getDFdatetime(df, dt_str, dt_format='%Y %m %d %H'):
    '''Format dataframe with datetime (year, month, day, hour) index column
    
    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame
    dt_str : list
        List of datetime strings to format and add
    dt_format : str
        Datetime string format. Default is "%Y %m %d %H".
    
    Returns
    -------
    df : pandas.DataFrame
        DataFrame with added datetime as index
    '''
    dates = [datetime.strptime(str(d), dt_format) for d in dt_str]
    df['time'] = dates
    df = df.set_index('time')
    df.drop(columns=df.columns[0], axis=1, inplace=True)  
    return df

#------------------------------------------------------------------------------
        
class TestGet(unittest.TestCase): 
    def testURL(self):
        '''Test URL retrieval'''
        l = lookup_table(['doi:10.22008/FK2/IW73UU'])
        self.assertTrue('10.22008/FK2' in list(l.values())[0])
    
#    def testAWSname(self):  
#        '''Test AWS names retrieval'''
#        n = aws_names()
#        self.assertIsInstance(n, list)
#        self.assertTrue('nuk_k_hour.csv' in n)
    
#    def testAWScsv(self):
#        '''Test AWS data retrieval'''
#        kan_b = aws_data('kan_b_hour.csv')
#        self.assertIsInstance(kan_b, pd.DataFrame)
        
#    def testWatsonHour(self):
#        '''Test Wason River discharge hourly data retrieval'''
#        wh = watson_discharge()
#        self.assertTrue(wh['Q']['2021-10-27 23:00:00']==5.48)
        
#    def testWatsonDaily(self):
#        '''Test Wason River discharge daily data retrieval'''
#        wd = watson_discharge(t='day')
#        self.assertTrue(wd['Q']['2009-09-04 00:00:00']==4.72)

    def testGetCLI(self):
        '''Test get_promice_data'''
        exit_status = os.system('get_promice_data -h')
        self.assertEqual(exit_status, 0)
            
if __name__ == "__main__": 
    unittest.main()



#!/usr/bin/env python
import logging, sys, os, unittest
import pandas as pd
import xarray as xr
from argparse import ArgumentParser
from pypromice.process.L1toL2 import correctPrecip
from pypromice.process.write import prepare_and_write
logger = logging.getLogger(__name__)

def parse_arguments_join():
    parser = ArgumentParser(description="AWS L2 joiner for merging together two L2 products, for example an L2 RAW and L2 TX data product. An hourly, daily and monthly L2 data product is outputted to the defined output path")
    parser.add_argument('-s', '--file1', type=str, required=True,
                        help='Path to source L2 file, which will be preferenced in merge process')
    parser.add_argument('-t', '--file2', type=str, required=True, 
                        help='Path to target L2 file, which will be used to fill gaps in merge process')
    parser.add_argument('-o', '--outpath', default=os.getcwd(), type=str, required=True, 
                        help='Path where to write output')
    parser.add_argument('-v', '--variables', default=None, type=str, required=False, 
    			 help='Path to variables look-up table .csv file for variable name retained'''),
    parser.add_argument('-m', '--metadata', default=None, type=str, required=False, 
    			 help='Path to metadata table .csv file for metadata information'''),
    args = parser.parse_args()
    return args

def loadArr(infile):
    if infile.split('.')[-1].lower() == 'csv':
        df = pd.read_csv(infile, index_col=0, parse_dates=True)
        ds = xr.Dataset.from_dataframe(df)  
    elif infile.split('.')[-1].lower() == 'nc':
        with xr.open_dataset(infile) as ds:
            ds.load()
        # Remove encoding attributes from NetCDF
        for varname in ds.variables:
            if ds[varname].encoding!={}:
                ds[varname].encoding = {}

    try:
        name = ds.attrs['station_id'] 
    except:
        name = infile.split('/')[-1].split('.')[0].split('_hour')[0].split('_10min')[0]
        ds.attrs['station_id'] = name
    if 'bedrock' in ds.attrs.keys():
        ds.attrs['bedrock'] = ds.attrs['bedrock'] == 'True'
    if 'number_of_booms' in ds.attrs.keys():
        ds.attrs['number_of_booms'] = int(ds.attrs['number_of_booms'])

    logger.info(f'{name} array loaded from {infile}')
    return ds, name
    

def join_l2(file1,file2,outpath,variables,metadata) -> xr.Dataset:
    logging.basicConfig(
        format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
        level=logging.INFO,
        stream=sys.stdout,
    )
    # Check files
    if os.path.isfile(file1) and os.path.isfile(file2): 

        # Load data arrays
        ds1, n1 = loadArr(file1)
        ds2, n2 = loadArr(file2)    	
        
        # Check stations match
        if n1.lower() == n2.lower():
            
        	# Merge arrays
            logger.info(f'Combining {file1} with {file2}...')
            name = n1
            all_ds = ds1.combine_first(ds2)
            
            # Re-calculate corrected precipitation
            if hasattr(all_ds, 'precip_u_cor'):
                if ~all_ds['precip_u_cor'].isnull().all():
                    all_ds['precip_u_cor'],  _ = correctPrecip(all_ds['precip_u'], 
                                                                all_ds['wspd_u'])
            if hasattr(all_ds, 'precip_l_cor'):
                if ~all_ds['precip_l_cor'].isnull().all():
                    all_ds['precip_l_cor'],  _ = correctPrecip(all_ds['precip_l'], 
                                                                all_ds['wspd_l'])                    
        else:
            logger.info(f'Mismatched station names {n1}, {n2}')
            exit()            
    
    elif os.path.isfile(file1):  
        ds1, name = loadArr(file1)
        logger.info(f'Only one file found {file1}...')
        all_ds = ds1  

    elif os.path.isfile(file2):
        ds2, name = loadArr(file2)
        logger.info(f'Only one file found {file2}...')
        all_ds = ds2  
    
    else:
        logger.info(f'Invalid files {file1}, {file2}')
        exit()

    all_ds.attrs['format'] = 'merged RAW and TX'

    # Resample to hourly, daily and monthly datasets and write to file
    prepare_and_write(all_ds, outpath, variables, metadata, resample = False)
    
    logger.info(f'Files saved to {os.path.join(outpath, name)}...')
    return all_ds

def main():
    args = parse_arguments_join()
    _ = join_l2(args.file1, args.file2, args.outpath, args.variables, args.metadata)
    
if __name__ == "__main__":  
    main()



#!/usr/bin/env python
from argparse import ArgumentParser
from configparser import ConfigParser
import os, imaplib, email, unittest
from glob import glob
from pypromice.tx import getMail


def parse_arguments_msg():
    parser = ArgumentParser(description="AWS message downloader")       
    parser.add_argument('-a', '--account', default=None, type=str, 
                        required=True, help='Email account .ini file')
    parser.add_argument('-p', '--password', default=None, type=str, 
                        required=True, help='Email credentials .ini file')                      
    parser.add_argument('-o', '--outpath', default='.', type=str, 
                        required=False, help='Path where to write output')    	
    parser.add_argument('-u', '--uid', default=None, type=str, 
                        required=False, help='Last uid, defined from files if not given')	
    parser.add_argument('-m', '--mailbox', default='"[Gmail]/All Mail"', 
                        type=str, required=False, help='Mailbox folder to collect messages from')        
    args = parser.parse_args()
    return args

def get_msg():
    args = parse_arguments_msg()

    # Set credential paths
    accounts_file = args.account
    credentials_file = args.password 
        
    # Set last aws uid path
    if args.uid is not None:
        last_uid = int(args.uid)
    else:
        last_uid=None
          
    
    # Set output file directory
    out_dir = args.outpath

    
    # Define accounts and credentials ini file paths
    accounts_ini = ConfigParser()
    accounts_ini.read_file(open(accounts_file))
    accounts_ini.read(credentials_file)
    
    # Get credentials
    account = accounts_ini.get('aws', 'account')
    server = accounts_ini.get('aws', 'server')
    port = accounts_ini.getint('aws', 'port')    
    password = accounts_ini.get('aws', 'password')
    if not password:
        password = input('password for AWS email account: ')
    print('AWS data from server %s, account %s' %(server, account))

    
    # Define last UID from file names
    if not os.path.exists(out_dir):
       os.mkdir(out_dir)
       if last_uid is None:
            last_uid=0
    else:
        if last_uid is None:
            fl = list(glob(os.path.join(out_dir,'*.msg')))
            fn = [int(f.split('_')[-1].split('.')[0]) for f in fl]
            last_uid = max(fn) 
        
    print('Fetching mail from uid %i' %last_uid)
   
    
    # Log in to email server
    mail_server = imaplib.IMAP4_SSL(server, port)
    typ, accountDetails = mail_server.login(account, password) 	
    if typ != 'OK':
        print('Not able to sign in!')
        raise

    # Grab new emails
    result, data = mail_server.select(mailbox='"[Gmail]/All Mail"', readonly=True)
    print('mailbox contains %s messages' %data[0])


    # Save mail to file
    for uid, mail in getMail(mail_server, last_uid=last_uid):
        message = email.message_from_string(mail)
        outfile = str(message['subject'])+'_'+str(uid)+'.msg'
          
        with open(os.path.join(out_dir, outfile), 'w') as out:
             out.write(mail)


    # Close mail server if open
    if 'mail_server' in locals():
        print(f'\nClosing {account}')
        mail_server.close()
        resp = mail_server.logout()
        assert resp[0].upper() == 'BYE'

    
    print('Finished')
        
if __name__ == "__main__":  
    get_msg()


#!/usr/bin/env python
"""
AWS Level 1 (L1) to Level 2 (L2) data processing
"""
import logging
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from pypromice.qc.github_data_issues import flagNAN, adjustTime, adjustData
from pypromice.qc.percentiles.outlier_detector import ThresholdBasedOutlierDetector
from pypromice.qc.persistence import persistence_qc
from pypromice.process.value_clipping import clip_values

__all__ = [
    "toL2",
]

logger = logging.getLogger(__name__)


def toL2(
    L1: xr.Dataset,
    vars_df: pd.DataFrame,
    data_flags_dir: Path,
    data_adjustments_dir: Path,
    T_0=273.15,
    ews=1013.246,
    ei0=6.1071,
    eps_overcast=1.0,
    eps_clear=9.36508e-6,
    emissivity=0.97,
) -> xr.Dataset:
    '''Process one Level 1 (L1) product to Level 2.
    In this step we do:
        - manual flagging and adjustments
        - automated QC: persistence, percentile
        - custom filter: gps_alt filter, NaN t_rad removed from dlr & ulr
        - smoothing of tilt and rot
        - calculation of rh with regards to ice in subfreezin conditions
        - calculation of cloud coverage
        - correction of dsr and usr for tilt
        - filtering of dsr based on a theoritical TOA irradiance and grazing light
        - calculation of albedo
        - calculation of directional wind speed

    Parameters
    ----------
    L1 : xarray.Dataset
        Level 1 dataset
    vars_df : pd.DataFrame
        Metadata dataframe
    T_0 : float
        Ice point temperature in K. The default is 273.15.
    ews : float
        Saturation pressure (normal atmosphere) at steam point temperature.
        The default is 1013.246.
    ei0 : float
        Saturation pressure (normal atmosphere) at ice-point temperature. The
        default is 6.1071.
    eps_overcast : int
        Cloud overcast. The default is 1..
    eps_clear : float
        Cloud clear. The default is 9.36508e-6.
    emissivity : float
        Emissivity. The default is 0.97.

    Returns
    -------
    ds : xarray.Dataset
        Level 2 dataset
    '''
    ds = L1.copy(deep=True)                                                    # Reassign dataset
    ds.attrs['level'] = 'L2'
    try:
        ds = adjustTime(ds, adj_dir=data_adjustments_dir.as_posix())       # Adjust time after a user-defined csv files
        ds = flagNAN(ds, flag_dir=data_flags_dir.as_posix())             # Flag NaNs after a user-defined csv files
        ds = adjustData(ds, adj_dir=data_adjustments_dir.as_posix())       # Adjust data after a user-defined csv files
    except Exception:
        logger.exception('Flagging and fixing failed:')

    ds = persistence_qc(ds)                                               # Flag and remove persistence outliers
    # if ds.attrs['format'] == 'TX':
    #     # TODO: The configuration should be provided explicitly
    #     outlier_detector = ThresholdBasedOutlierDetector.default()
    #     ds = outlier_detector.filter_data(ds)                                 # Flag and remove percentile outliers

    # filtering gps_lat, gps_lon and gps_alt based on the difference to a baseline elevation
    # right now baseline elevation is gapfilled monthly median elevation
    baseline_elevation = (ds.gps_alt.to_series().resample('MS').median()
                          .reindex(ds.time.to_series().index, method='nearest')
                          .ffill().bfill())
    mask = (np.abs(ds.gps_alt - baseline_elevation) < 100) | ds.gps_alt.isnull()
    ds[['gps_alt','gps_lon', 'gps_lat']] = ds[['gps_alt','gps_lon', 'gps_lat']].where(mask)

    # removing dlr and ulr that are missing t_rad
    # this is done now becasue t_rad can be filtered either manually or with persistence
    ds['dlr'] = ds.dlr.where(ds.t_rad.notnull())
    ds['ulr'] = ds.ulr.where(ds.t_rad.notnull())

    # calculating realtive humidity with regard to ice
    T_100 = _getTempK(T_0)
    ds['rh_u_wrt_ice_or_water'] = adjustHumidity(ds['rh_u'], ds['t_u'],
                                     T_0, T_100, ews, ei0)

    if ds.attrs['number_of_booms']==2:
        ds['rh_l_wrt_ice_or_water'] = adjustHumidity(ds['rh_l'], ds['t_l'],
                                         T_0, T_100, ews, ei0)

    if hasattr(ds,'t_i'):
        if ~ds['t_i'].isnull().all():
            ds['rh_i_wrt_ice_or_water'] = adjustHumidity(ds['rh_i'], ds['t_i'],
                                             T_0, T_100, ews, ei0)

    # Determiune cloud cover for on-ice stations
    cc = calcCloudCoverage(ds['t_u'], T_0, eps_overcast, eps_clear,        # Calculate cloud coverage
                           ds['dlr'], ds.attrs['station_id'])
    ds['cc'] = (('time'), cc.data)

    # Determine surface temperature
    ds['t_surf'] = calcSurfaceTemperature(T_0, ds['ulr'], ds['dlr'],           # Calculate surface temperature
                                          emissivity)
    if not ds.attrs['bedrock']:
        ds['t_surf'] = xr.where(ds['t_surf'] > 0, 0, ds['t_surf'])

    # Determine station position relative to sun
    doy = ds['time'].to_dataframe().index.dayofyear.values                     # Gather variables to calculate sun pos
    hour = ds['time'].to_dataframe().index.hour.values
    minute = ds['time'].to_dataframe().index.minute.values

    if hasattr(ds, 'latitude') and hasattr(ds, 'longitude'):
        lat = ds.attrs['latitude']                                             # TODO Why is mean GPS lat lon not preferred for calcs?
        lon = ds.attrs['longitude']
    else:
        lat = ds['gps_lat'].mean()
        lon = ds['gps_lon'].mean()

    # smoothing tilt and rot
    ds['tilt_x'] = smoothTilt(ds['tilt_x'])
    ds['tilt_y'] = smoothTilt(ds['tilt_y'])
    ds['rot'] = smoothRot(ds['rot'])

    deg2rad, rad2deg = _getRotation()                                          # Get degree-radian conversions
    phi_sensor_rad, theta_sensor_rad = calcTilt(ds['tilt_x'], ds['tilt_y'],    # Calculate station tilt
                                                deg2rad)

    Declination_rad = calcDeclination(doy, hour, minute)                       # Calculate declination
    HourAngle_rad = calcHourAngle(hour, minute, lon)                           # Calculate hour angle
    ZenithAngle_rad, ZenithAngle_deg = calcZenith(lat, Declination_rad,        # Calculate zenith
                                                  HourAngle_rad, deg2rad,
                                                  rad2deg)


    # Correct Downwelling shortwave radiation
    DifFrac = 0.2 + 0.8 * cc
    CorFac_all = calcCorrectionFactor(Declination_rad, phi_sensor_rad,         # Calculate correction
                                      theta_sensor_rad, HourAngle_rad,
                                      ZenithAngle_rad, ZenithAngle_deg,
                                      lat, DifFrac, deg2rad)
    CorFac_all = xr.where(ds['cc'].notnull(), CorFac_all, 1)
    ds['dsr_cor'] = ds['dsr'].copy(deep=True) * CorFac_all                     # Apply correction

    AngleDif_deg = calcAngleDiff(ZenithAngle_rad, HourAngle_rad,               # Calculate angle between sun and sensor
                                 phi_sensor_rad, theta_sensor_rad)

    ds['albedo'], OKalbedos = calcAlbedo(ds['usr'], ds['dsr_cor'],             # Determine albedo
                              AngleDif_deg, ZenithAngle_deg)

    # Correct upwelling and downwelling shortwave radiation
    sunonlowerdome =(AngleDif_deg >= 90) & (ZenithAngle_deg <= 90)             # Determine when sun is in FOV of lower sensor, assuming sensor measures only diffuse radiation
    ds['dsr_cor'] = ds['dsr_cor'].where(~sunonlowerdome,
                                        other=ds['dsr'] / DifFrac)             # Apply to downwelling
    ds['usr_cor'] = ds['usr'].copy(deep=True)
    ds['usr_cor'] = ds['usr_cor'].where(~sunonlowerdome,
                                        other=ds['albedo'] * ds['dsr'] / DifFrac) # Apply to upwelling
    bad = (ZenithAngle_deg > 95) | (ds['dsr_cor'] <= 0) | (ds['usr_cor'] <= 0) # Set to zero for solar zenith angles larger than 95 deg or either values are (less than) zero
    ds['dsr_cor'][bad] = 0
    ds['usr_cor'][bad] = 0
    ds['dsr_cor'] = ds['usr_cor'].copy(deep=True) / ds['albedo']               # Correct DWR using more reliable USWR when sun not in sight of upper sensor
    ds['albedo'] = ds['albedo'].where(OKalbedos)                               #TODO remove?

    # Remove data where TOA shortwave radiation invalid
    isr_toa = calcTOA(ZenithAngle_deg, ZenithAngle_rad)                        # Calculate TOA shortwave radiation
    TOA_crit_nopass = (ds['dsr_cor'] > (0.9 * isr_toa + 10))                   # Determine filter
    ds['dsr_cor'][TOA_crit_nopass] = np.nan                                    # Apply filter and interpolate
    ds['usr_cor'][TOA_crit_nopass] = np.nan

    ds['dsr_cor'] = ds.dsr_cor.where(ds.dsr.notnull())
    ds['usr_cor'] = ds.usr_cor.where(ds.usr.notnull())
    # # Check sun position
    # sundown = ZenithAngle_deg >= 90
    # _checkSunPos(ds, OKalbedos, sundown, sunonlowerdome, TOA_crit_nopass)

    if hasattr(ds, 'correct_precip'):                                          # Correct precipitation
        precip_flag=ds.attrs['correct_precip']
    else:
        precip_flag=True
    if ~ds['precip_u'].isnull().all() and precip_flag:
        ds['precip_u_cor'], ds['precip_u_rate'] = correctPrecip(ds['precip_u'],
                                                                ds['wspd_u'])
    if ds.attrs['number_of_booms']==2:
        if ~ds['precip_l'].isnull().all() and precip_flag:                     # Correct precipitation
            ds['precip_l_cor'], ds['precip_l_rate']= correctPrecip(ds['precip_l'],
                                                                   ds['wspd_l'])

    get_directional_wind_speed(ds)                                            # Get directional wind speed

    ds = clip_values(ds, vars_df)
    return ds

def get_directional_wind_speed(ds: xr.Dataset) -> xr.Dataset:
    """
    Calculate directional wind speed from wind speed and direction and mutates the dataset
    """

    ds['wdir_u'] = ds['wdir_u'].where(ds['wspd_u'] != 0)
    ds['wspd_x_u'], ds['wspd_y_u'] = calcDirWindSpeeds(ds['wspd_u'], ds['wdir_u'])

    if ds.attrs['number_of_booms']==2:
        ds['wdir_l'] = ds['wdir_l'].where(ds['wspd_l'] != 0)
        ds['wspd_x_l'], ds['wspd_y_l'] = calcDirWindSpeeds(ds['wspd_l'], ds['wdir_l'])

    if hasattr(ds, 'wdir_i'):
        if ~ds['wdir_i'].isnull().all() and ~ds['wspd_i'].isnull().all():
            ds['wdir_i'] = ds['wdir_i'].where(ds['wspd_i'] != 0)
            ds['wspd_x_i'], ds['wspd_y_i'] = calcDirWindSpeeds(ds['wspd_i'], ds['wdir_i'])
    return ds


def calcDirWindSpeeds(wspd, wdir, deg2rad=np.pi/180):
    '''Calculate directional wind speed from wind speed and direction

    Parameters
    ----------
    wspd : xr.Dataarray
        Wind speed data array
    wdir : xr.Dataarray
        Wind direction data array
    deg2rad : float
        Degree to radians coefficient. The default is np.pi/180

    Returns
    -------
    wspd_x : xr.Dataarray
        Wind speed in X direction
    wspd_y : xr.Datarray
        Wind speed in Y direction
    '''
    wspd_x = wspd * np.sin(wdir * deg2rad)
    wspd_y = wspd * np.cos(wdir * deg2rad)
    return wspd_x, wspd_y


def calcCloudCoverage(T, T_0, eps_overcast, eps_clear, dlr, station_id):
    '''Calculate cloud cover from T and T_0

    Parameters
    ----------
    T : xarray.DataArray
        Air temperature 1
    T_0 : xarray.DataArray
        Air temperature 0
    eps_overcast : int
        Cloud overcast assumption, from Swinbank (1963)
    eps_clear : int
        Cloud clear assumption, from Swinbank (1963)
    dlr : xarray.DataArray
        Downwelling longwave radiation, with array of same length as T and T_0
    station_id : str
        Station ID string, for special cases at selected stations where cloud
        overcast and cloud clear assumptions are pre-defined. Currently
        KAN_M and KAN_U are special cases, but this will need to be done for
        all stations eventually

    Returns
    -------
    cc : xarray.DataArray
        Cloud cover data array
    '''
    if station_id == 'KAN_M':
       LR_overcast = 315 + 4*T
       LR_clear = 30 + 4.6e-13 * (T + T_0)**6
    elif station_id == 'KAN_U':
       LR_overcast = 305 + 4*T
       LR_clear = 220 + 3.5*T
    else:
       LR_overcast = eps_overcast * 5.67e-8 *(T + T_0)**4
       LR_clear = eps_clear * 5.67e-8 * (T + T_0)**6
    cc = (dlr - LR_clear) / (LR_overcast - LR_clear)
    cc[cc > 1] = 1
    cc[cc < 0] = 0
    return cc


def calcSurfaceTemperature(T_0, ulr, dlr, emissivity):
    '''Calculate surface temperature from air temperature, upwelling and
    downwelling radiation and emissivity

    Parameters
    ----------
    T_0 : xarray.DataArray
        Air temperature
    ulr : xarray.DataArray
        Upwelling longwave radiation
    dlr : xarray.DataArray
        Downwelling longwave radiation
    emissivity : int
        Assumed emissivity

    Returns
    -------
    xarray.DataArray
        Calculated surface temperature
    '''
    t_surf = ((ulr - (1 - emissivity) * dlr) / emissivity / 5.67e-8)**0.25 - T_0
    return t_surf


def smoothTilt(da: xr.DataArray, threshold=0.2):
    '''Smooth the station tilt

    Parameters
    ----------
    da : xarray.DataArray
        either X or Y tilt inclinometer measurements
    threshold : float
        threshold used in a standrad.-deviation based filter

    Returns
    -------
    xarray.DataArray
        either X or Y smoothed tilt inclinometer measurements
    '''
    # we calculate the moving standard deviation over a 3-day sliding window
    # hourly resampling is necessary to make sure the same threshold can be used
    # for 10 min and hourly data
    moving_std_gap_filled = da.to_series().resample('h').median().rolling(
                    3*24, center=True, min_periods=2
                    ).std().reindex(da.time, method='bfill').values
    # we select the good timestamps and gapfill assuming that
    # - when tilt goes missing the last available value is used
    # - when tilt is not available for the very first time steps, the first
    #   good value is used for backfill
    return da.where(
                moving_std_gap_filled < threshold
                ).ffill(dim='time').bfill(dim='time')


def smoothRot(da: xr.DataArray, threshold=4):
    '''Smooth the station rotation

    Parameters
    ----------
    da : xarray.DataArray
        rotation measurements from inclinometer
    threshold : float
        threshold used in a standrad-deviation based filter

    Returns
    -------
    xarray.DataArray
        smoothed rotation measurements from inclinometer
    '''
    moving_std_gap_filled = da.to_series().resample('h').median().rolling(
                    3*24, center=True, min_periods=2
                    ).std().reindex(da.time, method='bfill').values
    # same as for tilt with, in addition:
    #     - a resampling to daily values
    #     - a two week median smoothing
    #     - a resampling from these daily values to the original temporal resolution
    return ('time', (da.where(moving_std_gap_filled <4).ffill(dim='time')
            .to_series().resample('D').median()
            .rolling(7*2,center=True,min_periods=2).median()
            .reindex(da.time, method='bfill').values
            ))


def calcTilt(tilt_x, tilt_y, deg2rad):
    '''Calculate station tilt

    Parameters
    ----------
    tilt_x : xarray.DataArray
        X tilt inclinometer measurements
    tilt_y : xarray.DataArray
        Y tilt inclinometer measurements
    deg2rad : float
        Degrees to radians conversion

    Returns
    -------
    phi_sensor_rad : xarray.DataArray
        Spherical tilt coordinates
    theta_sensor_rad : xarray.DataArray
        Total tilt of sensor, where 0 is horizontal
    '''
    # Tilt as radians
    tx = tilt_x * deg2rad
    ty = tilt_y * deg2rad

    # Calculate cartesian coordinates
    X = np.sin(tx) * np.cos(tx) * np.sin(ty)**2 + np.sin(tx) * np.cos(ty)**2
    Y = np.sin(ty) * np.cos(ty) * np.sin(tx)**2 + np.sin(ty) * np.cos(tx)**2
    Z = np.cos(tx) * np.cos(ty) + np.sin(tx)**2 * np.sin(ty)**2

    # Calculate spherical coordinates
    phi_sensor_rad = -np.pi /2 - np.arctan(Y/X)
    phi_sensor_rad[X > 0] += np.pi
    phi_sensor_rad[(X == 0) & (Y < 0)] = np.pi
    phi_sensor_rad[(X == 0) & (Y == 0)] = 0
    phi_sensor_rad[phi_sensor_rad < 0] += 2*np.pi

    # Total tilt of the sensor, i.e. 0 when horizontal
    theta_sensor_rad = np.arccos(Z / (X**2 + Y**2 + Z**2)**0.5)
    # phi_sensor_deg = phi_sensor_rad * rad2deg                                #TODO take these out if not needed
    # theta_sensor_deg = theta_sensor_rad * rad2deg
    return phi_sensor_rad, theta_sensor_rad


def adjustHumidity(rh, T, T_0, T_100, ews, ei0):                        #TODO figure out if T replicate is needed
    '''Adjust relative humidity so that values are given with respect to
    saturation over ice in subfreezing conditions, and with respect to
    saturation over water (as given by the instrument) above the melting
    point temperature. Saturation water vapors are calculated after
    Groff & Gratch method.

    Parameters
    ----------
    rh : xarray.DataArray
        Relative humidity
    T : xarray.DataArray
        Air temperature
    T_0 : float
        Ice point temperature in K
    T_100 : float
        Steam point temperature in K
    ews : float
        Saturation pressure (normal atmosphere) at steam point temperature
    ei0 : float
        Saturation pressure (normal atmosphere) at ice-point temperature

    Returns
    -------
    rh_wrt_ice_or_water : xarray.DataArray
        Corrected relative humidity
    '''
    # Convert to hPa (Groff & Gratch)
    e_s_wtr = 10**(-7.90298 * (T_100 / (T + T_0) - 1)
                   + 5.02808 * np.log10(T_100 / (T + T_0))
                   - 1.3816E-7 * (10**(11.344 * (1 - (T + T_0) / T_100)) - 1)
                   + 8.1328E-3 * (10**(-3.49149 * (T_100/(T + T_0) - 1)) -1)
                   + np.log10(ews))
    e_s_ice = 10**(-9.09718 * (T_0 / (T + T_0) - 1)
                   - 3.56654 * np.log10(T_0 / (T + T_0))
                   + 0.876793 * (1 - (T + T_0) / T_0)
                   + np.log10(ei0))

    # Define freezing point. Why > -100?
    freezing = (T < 0) & (T > -100).values

    # Set to Groff & Gratch values when freezing, otherwise just rh
    rh_wrt_ice_or_water = rh.where(~freezing, other = rh*(e_s_wtr / e_s_ice))
    return rh_wrt_ice_or_water


def correctPrecip(precip, wspd):
    '''Correct precipitation with the undercatch correction method used in
    Yang et al. (1999) and Box et al. (2022), based on Goodison et al. (1998)

    Yang, D., Ishida, S., Goodison, B. E., and Gunther, T.: Bias correction of
    daily precipitation measurements for Greenland,
    https://doi.org/10.1029/1998jd200110, 1999.

    Box, J., Wehrle, A., van As, D., Fausto, R., Kjeldsen, K., Dachauer, A.,
    Ahlstrom, A. P., and Picard, G.: Greenland Ice Sheet rainfall, heat and
    albedo feedback imapacts from the Mid-August 2021 atmospheric river,
    Geophys. Res. Lett. 49 (11), e2021GL097356,
    https://doi.org/10.1029/2021GL097356, 2022.

    Goodison, B. E., Louie, P. Y. T., and Yang, D.: Solid Precipitation
    Measurement Intercomparison, WMO, 1998

    Parameters
    ----------
    precip : xarray.DataArray
        Cumulative precipitation measurements
    wspd : xarray.DataArray
        Wind speed measurements

    Returns
    -------
    precip_cor : xarray.DataArray
        Cumulative precipitation corrected
    precip_rate : xarray.DataArray
        Precipitation rate corrected
    '''
    # Calculate undercatch correction factor
    corr=100/(100.00-4.37*wspd+0.35*wspd*wspd)

    # Fix all values below 1.02 to 1.02
    corr = corr.where(corr>1.02, other=1.02)

    # Fill nan values in precip with preceding value
    precip = precip.ffill(dim='time')

    # Calculate precipitation rate
    precip_rate = precip.diff(dim='time', n=1)

    # Apply correction to rate
    precip_rate = precip_rate*corr

    # Flag rain bucket reset
    precip_rate = precip_rate.where(precip_rate>-0.01, other=np.nan)
    b = precip_rate.to_dataframe('precip_flag').notna().to_xarray()

    # Get corrected cumulative precipitation, reset if rain bucket flag
    precip_cor = precip_rate.cumsum()-precip_rate.cumsum().where(~b['precip_flag']).ffill(dim='time').fillna(0).astype(float)

    return precip_cor, precip_rate


def calcDeclination(doy, hour, minute):
    '''Calculate sun declination based on time

    Parameters
    ----------
    doy : int
        Day of year
    hour : int
        Hour of day
    minute : int
        Minute of hour

    Returns
    -------
    float
        Sun declination
    '''
    d0_rad = 2 * np.pi * (doy + (hour + minute / 60) / 24 -1) / 365
    return np.arcsin(0.006918 - 0.399912
                     * np.cos(d0_rad) + 0.070257
                     * np.sin(d0_rad) - 0.006758
                     * np.cos(2 * d0_rad) + 0.000907
                     * np.sin(2 * d0_rad) - 0.002697
                     * np.cos(3 * d0_rad) + 0.00148
                     * np.sin(3 * d0_rad))

def calcHourAngle(hour, minute, lon):
    '''Calculate hour angle of sun based on time and longitude. Make sure that
    time is set to UTC and longitude is positive when west. Hour angle should
    be 0 at noon

    Parameters
    ----------
    hour : int
        Hour of day
    minute : int
        Minute of hour
    lon : float
        Longitude

    Returns
    -------
    float
        Hour angle of sun
    '''
    return 2 * np.pi * (((hour + minute / 60) / 24 - 0.5) - lon/360)
     # ; - 15.*timezone/360.)


def calcDirectionDeg(HourAngle_rad):                                          #TODO remove if not plan to use this
    '''Calculate sun direction as degrees. This is an alternative to
    _calcHourAngle that is currently not implemented into the offical L0>>L3
    workflow. Here, 180 degrees is at noon (NH), as opposed to HourAngle

    Parameters
    ----------
    HourAngle_rad : float
        Sun hour angle in radians

    Returns
    -------
    DirectionSun_deg
        Sun direction in degrees
    '''
    DirectionSun_deg = HourAngle_rad * 180/np.pi - 180
    DirectionSun_deg[DirectionSun_deg < 0] += 360
    DirectionSun_deg[DirectionSun_deg < 0] += 360
    return DirectionSun_deg

def calcZenith(lat, Declination_rad, HourAngle_rad, deg2rad, rad2deg):
    '''Calculate sun zenith in radians and degrees

    Parameters
    ----------
    lat : float
        Latitude
    Declination_Rad : float
        Sun declination in radians
    HourAngle_rad : float
        Sun hour angle in radians
    deg2rad : float
        Degrees to radians conversion
    rad2deg : float
        Radians to degrees conversion

    Returns
    -------
    ZenithAngle_rad : float
        Zenith angle in radians
    ZenithAngle_deg : float
        Zenith angle in degrees
    '''
    ZenithAngle_rad = np.arccos(np.cos(lat * deg2rad)
                                * np.cos(Declination_rad)
                                * np.cos(HourAngle_rad)
                                + np.sin(lat * deg2rad)
                                * np.sin(Declination_rad))

    ZenithAngle_deg = ZenithAngle_rad * rad2deg
    return ZenithAngle_rad, ZenithAngle_deg


def calcAngleDiff(ZenithAngle_rad, HourAngle_rad, phi_sensor_rad,
                  theta_sensor_rad):
    '''Calculate angle between sun and upper sensor (to determine when sun is
    in sight of upper sensor

    Parameters
    ----------
    ZenithAngle_rad : float
        Zenith angle in radians
    HourAngle_rad : float
        Sun hour angle in radians
    phi_sensor_rad : xarray.DataArray
        Spherical tilt coordinates
    theta_sensor_rad : xarray.DataArray
        Total tilt of sensor, where 0 is horizontal

    Returns
    -------
    float
        Angle between sun and sensor
    '''
    return 180 / np.pi * np.arccos(np.sin(ZenithAngle_rad)
                                   * np.cos(HourAngle_rad + np.pi)
                                   * np.sin(theta_sensor_rad)
                                   * np.cos(phi_sensor_rad)
                                   + np.sin(ZenithAngle_rad)
                                   * np.sin(HourAngle_rad + np.pi)
                                   * np.sin(theta_sensor_rad)
                                   * np.sin(phi_sensor_rad)
                                   + np.cos(ZenithAngle_rad)
                                   * np.cos(theta_sensor_rad))

def calcAlbedo(usr, dsr_cor, AngleDif_deg, ZenithAngle_deg):
    '''Calculate surface albedo based on upwelling and downwelling shorwave
    flux, the angle between the sun and sensor, and the sun zenith

    Parameters
    ----------
    usr : xarray.DataArray
        Upwelling shortwave radiation
    dsr_cor : xarray.DataArray
        Downwelling shortwave radiation corrected
    AngleDif_def : float
        Angle between sun and sensor in degrees
    ZenithAngle_deg: float
        Zenith angle in degrees

    Returns
    -------
    albedo : xarray.DataArray
        Derived albedo
    OKalbedos : xarray.DataArray
        Valid albedo measurements
    '''
    albedo = usr / dsr_cor

    # NaN bad data
    OKalbedos = (AngleDif_deg < 70) & (ZenithAngle_deg < 70) & (albedo < 1) & (albedo > 0)
    albedo[~OKalbedos] = np.nan

    # Interpolate all. Note "use_coordinate=False" is used here to force
    # comparison against the GDL code when that is run with *only* a TX file.
    # Should eventually set to default (True) and interpolate based on time,
    # not index.
    albedo = albedo.interpolate_na(dim='time', use_coordinate=False)
    albedo = albedo.ffill(dim='time').bfill(dim='time')                        #TODO remove this line and one above?
    return albedo, OKalbedos


def calcTOA(ZenithAngle_deg, ZenithAngle_rad):
    '''Calculate incoming shortwave radiation at the top of the atmosphere,
    accounting for sunset periods

    Parameters
    ----------
    ZenithAngle_deg : float
        Zenith angle in degrees
    ZenithAngle_rad : float
        Zenith angle in radians

    Returns
    -------
    isr_toa : float
        Incoming shortwave radiation at the top of the atmosphere
    '''
    sundown = ZenithAngle_deg >= 90

    # Incoming shortware radiation at the top of the atmosphere
    isr_toa = 1372 * np.cos(ZenithAngle_rad)
    isr_toa[sundown] = 0
    return isr_toa

def calcCorrectionFactor(Declination_rad, phi_sensor_rad, theta_sensor_rad,
                          HourAngle_rad, ZenithAngle_rad, ZenithAngle_deg,
                          lat, DifFrac, deg2rad):
    '''Calculate correction factor for direct beam radiation, as described
    here: http://solardat.uoregon.edu/SolarRadiationBasics.html

    Offset correction (where solar zenith angles larger than 110 degrees) not
    implemented as it should not improve the accuracy of well-calibrated
    instruments. It would go something like this:
    ds['dsr'] = ds['dsr'] - ds['dwr_offset']
    SRout = SRout - SRout_offset

    Parameters
    ----------
    Declination_rad : float
        Declination in radians
    phi_sensor_rad : xarray.DataArray
        Spherical tilt coordinates
    theta_sensor_rad : xarray.DataArray
        Total tilt of sensor, where 0 is horizontal
    HourAngle_rad : float
        Sun hour angle in radians
    ZenithAngle_rad : float
        Zenith angle in radians
    ZenithAngle_deg : float
        Zenith Angle in degrees
    lat :  float
        Latitude
    DifFrac : float
        Fractional cloud cover
    deg2rad : float
        Degrees to radians conversion

    Returns
    -------
    CorFac_all : xarray.DataArray
        Correction factor
    '''
    CorFac = np.sin(Declination_rad) * np.sin(lat * deg2rad) \
        * np.cos(theta_sensor_rad) \
        - np.sin(Declination_rad) \
        * np.cos(lat * deg2rad) \
        * np.sin(theta_sensor_rad) \
        * np.cos(phi_sensor_rad + np.pi) \
        + np.cos(Declination_rad) \
        * np.cos(lat * deg2rad) \
        * np.cos(theta_sensor_rad) \
        * np.cos(HourAngle_rad) \
        + np.cos(Declination_rad) \
        * np.sin(lat * deg2rad) \
        * np.sin(theta_sensor_rad) \
        * np.cos(phi_sensor_rad + np.pi) \
        * np.cos(HourAngle_rad) \
        + np.cos(Declination_rad) \
        * np.sin(theta_sensor_rad) \
        * np.sin(phi_sensor_rad + np.pi) \
        * np.sin(HourAngle_rad) \

    CorFac = np.cos(ZenithAngle_rad) / CorFac
    # sun out of field of view upper sensor
    CorFac[(CorFac < 0) | (ZenithAngle_deg > 90)] = 1

    # Calculating ds['dsr'] over a horizontal surface corrected for station/sensor tilt
    CorFac_all = CorFac / (1 - DifFrac + CorFac * DifFrac)

    return CorFac_all


def _checkSunPos(ds, OKalbedos, sundown, sunonlowerdome, TOA_crit_nopass):
    '''Check sun position

    Parameters
    ----------
    ds : xarray.Dataset
        Data set
    OKalbedos : xarray.DataArray
        Valid measurements flag
    sundown : xarray.DataArray
        Sun below horizon flag
    sunonlowerdome : xarray.DataArray
        Sun in view of lower sensor flag
    TOA_crit_nopass : xarray.DataArray
        Top-of-Atmosphere flag
    '''
    valid = (~(ds['dsr_cor'].isnull())).sum()
    print('Sun in view of upper sensor / workable albedos:', OKalbedos.sum().values,
          (100*OKalbedos.sum()/valid).round().values, "%")
    print('Sun below horizon:', sundown.sum(),
          (100*sundown.sum()/valid).round().values, "%")
    print('Sun in view of lower sensor:', sunonlowerdome.sum().values,
          (100*sunonlowerdome.sum()/valid).round().values, "%")
    print('Spikes removed using TOA criteria:', TOA_crit_nopass.sum().values,
          (100*TOA_crit_nopass.sum()/valid).round().values, "%")
    print('Mean net SR change by corrections:',
          (ds['dsr_cor']-ds['usr_cor']-ds['dsr']+ds['usr']).sum().values/valid.values,
          "W/m2")

def _getTempK(T_0):                                                            #TODO same as L2toL3._getTempK()
    '''Return steam point temperature in Kelvins

    Parameters
    ----------
    T_0 : float
        Ice point temperature in K

    Returns
    -------
    float
        Steam point temperature in K'''
    return T_0+100


def _getRotation():                                                            #TODO same as L2toL3._getRotation()
    '''Return degrees-to-radians and radians-to-degrees'''
    deg2rad = np.pi / 180
    rad2deg = 1 / deg2rad
    return deg2rad, rad2deg


if __name__ == "__main__":
    # unittest.main()
    pass



import sys
from datetime import datetime

import pandas as pd

from pypromice.process import AWS
from pathlib import Path
import logging
from pypromice.qc.github_data_issues import adjustTime, flagNAN, adjustData


# %%
logger = logging.getLogger("ComputeThreshold")


# %%
def compute_all_thresholds(
    station_thresholds_root: Path,
    thresholds_output_path: Path,
    aws_l0_repo_path: Path,
    start_time: datetime,
    end_time: datetime,
):
    logger.info("Computing all thresholds for stations available in the L0 repository")
    logger.info(f"station_thresholds_root: {station_thresholds_root}")
    logger.info(f"thresholds_output_path:  {thresholds_output_path}")
    logger.info(f"aws_l0_repo_path:        {aws_l0_repo_path}")
    logger.info(f"start_time:              {start_time}")
    logger.info(f"end_time:                {end_time}")

    station_thresholds_root.mkdir(parents=True, exist_ok=True)

    # %%
    output_paths = []
    for config_path in aws_l0_repo_path.glob("raw/config/*.toml"):
        stid = config_path.stem

        logger.info(f"Processing {stid}")
        data_path = aws_l0_repo_path.joinpath("raw", stid)
        output_path = station_thresholds_root.joinpath(f"{stid}.csv")
        try:
            if not output_path.exists():
                threshold = find_thresholds(
                    stid,
                    config_path,
                    data_path,
                    start_time,
                    end_time,
                )
                threshold.to_csv(
                    path_or_buf=output_path, index=False, float_format="{:.2f}".format
                )
            output_paths.append(output_path)
        except Exception:
            logger.exception(f"Failed processing {stid}")
            continue

    logger.info("Merge threshold files")
    pd.concat(pd.read_csv(p) for p in output_paths).to_csv(
        thresholds_output_path, index=False, float_format="{:.2f}".format
    )


def find_thresholds(
    stid: str,
    config_path: Path,
    data_path: Path,
    start_time: datetime,
    end_time: datetime,
) -> pd.DataFrame:
    """
    Compute variable threshold for a station using historical distribution quantiles.

    Parameters
    ----------
    stid
    config_path
    data_path
    start_time
    end_time

    Returns
    -------
    Upper and lower thresholds for a set of variables and seasons


    """
    stid_logger = logger.getChild(stid)
    # %%

    stid_logger.info("Read AWS data and get L1")
    aws = AWS(config_file=config_path.as_posix(), inpath=data_path.as_posix())
    aws.getL1()

    # %%
    stid_logger.info("Apply QC filters on data")
    ds = aws.L1A.copy(deep=True)  # Reassign dataset
    ds = adjustTime(ds)  # Adjust time after a user-defined csv files
    ds = flagNAN(ds)  # Flag NaNs after a user-defined csv files
    ds = adjustData(ds)

    # %%
    stid_logger.info("Determine thresholds")
    df = (
        ds[["rh_u", "wspd_u", "p_u", "t_u"]]
        .to_pandas()
        .loc[start_time:end_time]
        .assign(season=lambda df: (df.index.month // 3) % 4)
    )

    threshold_rows = []

    # Pressure
    p_lo, p_hi = df["p_u"].quantile([0.005, 0.995]) + [-12, 12]
    threshold_rows.append(
        dict(
            stid=stid,
            variable_pattern="p_[ul]",
            lo=p_lo,
            hi=p_hi,
        )
    )
    threshold_rows.append(
        dict(
            stid=stid,
            variable_pattern="p_i",
            lo=p_lo - 1000,
            hi=p_hi - 1000,
        )
    )

    # Wind speed
    lo, hi = df["wspd_u"].quantile([0.005, 0.995]) + [-12, 12]
    threshold_rows.append(
        dict(
            stid=stid,
            variable_pattern="wspd_[uli]",
            lo=lo,
            hi=hi,
        )
    )

    # Temperature
    season_map = ["winter", "spring", "summer", "fall"]
    for season_index, season_df in df[["t_u", "season"]].groupby(
        (df.index.month // 3) % 4
    ):
        lo, hi = season_df.quantile([0.005, 0.995])["t_u"] + [-9, 9]

        threshold_rows.append(
            dict(
                stid=stid,
                variable_pattern="t_[uli]",
                season=season_map[season_index],
                lo=lo,
                hi=hi,
            )
        )

    threshold = pd.DataFrame(threshold_rows)
    stid_logger.info(threshold)
    return threshold
    # %%


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--l0",
        required=True,
        type=Path,
        help="L0 repository root path",
    )
    parser.add_argument(
        "--thresholds_output_path",
        "-o",
        default=Path(__file__).parent.joinpath("thresholds.csv"),
        type=Path,
        help="Output csv file with thresholds for all stations",
    )
    parser.add_argument(
        "--station_thresholds_root",
        "--str",
        default=Path(__file__).parent.joinpath("station_thresholds"),
        type=Path,
        help="Directory containing threshold files for the individual stations",
    )
    parser.add_argument(
        "--start_time",
        default="2000-01-01",
        help="Start time for data series. Format: %Y-%m-%d",
    )
    parser.add_argument(
        "--end_time",
        default="2023-10-01",
        help="End time for data series. Format: %Y-%m-%d",
    )
    args = parser.parse_args()

    logging.basicConfig(
        format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
        level=logging.INFO,
        stream=sys.stdout,
    )

    thresholds_output_path = args.thresholds_output_path
    station_thresholds_root = args.station_thresholds_root

    start_time = datetime.strptime(args.start_time, "%Y-%m-%d")
    end_time = datetime.strptime(args.end_time, "%Y-%m-%d")
    aws_l0_repo_path = args.l0

    compute_all_thresholds(
        station_thresholds_root=station_thresholds_root,
        thresholds_output_path=thresholds_output_path,
        aws_l0_repo_path=aws_l0_repo_path,
        start_time=start_time,
        end_time=end_time,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module containing all the functions needed to prepare and AWS data
"""
import datetime
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd
from pypromice.process.resample import resample_dataset
import pypromice.resources

logger = logging.getLogger(__name__)


def prepare_and_write(
    dataset,
        output_path: Path | str,
        vars_df=None,
        meta_dict=None,
        time="60min",
        resample=True,
        nc_compression:bool=False,
):
    """Prepare data with resampling, formating and metadata population; then
    write data to .nc and .csv hourly and daily files

    Parameters
    ----------
    dataset : xarray.Dataset
        Dataset to write to file
    output_path : Path|str
        Output directory
    vars_df : pandas.DataFrame
        Variables look-up table dataframe
    meta_dict : dictionary
        Metadata dictionary to write to dataset
    time : str
        Resampling interval for output dataset
    """
    # Resample dataset
    if isinstance(output_path, str):
        output_path = Path(output_path)

    if resample:
        d2 = resample_dataset(dataset, time)
        logger.info("Resampling to " + str(time))
        if len(d2.time) == 1:
            logger.warning(
                "Output of resample has length 1. Not enough data to calculate daily/monthly average."
            )
            return None
    else:
        d2 = dataset.copy()

    # Reformat time
    d2 = reformat_time(d2)

    # finding station/site name
    if "station_id" in d2.attrs.keys():
        name = d2.attrs["station_id"]
    else:
        name = d2.attrs["site_id"]

    # Reformat longitude (to negative values)
    if "gps_lon" in d2.keys():
        d2 = reformat_lon(d2)
    else:
        logger.info("%s does not have gps_lon" % name)

    # Add variable attributes and metadata
    if vars_df is None:
        vars_df = pypromice.resources.load_variables()
    if meta_dict is None:
        meta_dict = pypromice.resources.load_metadata()

    d2 = addVars(d2, vars_df)
    d2 = addMeta(d2, meta_dict)

    # Round all values to specified decimals places
    d2 = roundValues(d2, vars_df)

    # Get variable names to write out
    if "site_id" in d2.attrs.keys():
        remove_nan_fields = True
    else:
        remove_nan_fields = False
    col_names = getColNames(vars_df, d2, remove_nan_fields=remove_nan_fields)

    # Define filename based on resample rate
    t = int(pd.Timedelta((d2["time"][1] - d2["time"][0]).values).total_seconds())

    # Create out directory
    output_dir = output_path / name
    output_dir.mkdir(exist_ok=True, parents=True)

    if t == 600:
        out_csv = output_dir / f"{name}_10min.csv"
        out_nc = output_dir / f"{name}_10min.nc"
    elif t == 3600:
        out_csv = output_dir / f"{name}_hour.csv"
        out_nc = output_dir / f"{name}_hour.nc"
    elif t == 86400:
        # removing instantaneous values from daily and monthly files
        for v in col_names:
            if v in ['p_i', 't_i', 'rh_i', 'wspd_i', 'wdir_i', 'wspd_x_i', 'wspd_y_i']:
                col_names.remove(v)
        out_csv = output_dir / f"{name}_day.csv"
        out_nc = output_dir / f"{name}_day.nc"
    else:
        # removing instantaneous values from daily and monthly files
        for v in col_names:
            if v in ['p_i', 't_i', 'rh_i', 'wspd_i', 'wdir_i', 'wspd_x_i', 'wspd_y_i']:
                col_names.remove(v)
        out_csv = output_dir / f"{name}_month.csv"
        out_nc = output_dir / f"{name}_month.nc"

    # Write to csv file
    logger.info("Writing to files...")
    writeCSV(out_csv, d2, col_names)

    # Write to netcdf file
    writeNC(out_nc, d2, col_names, compression=nc_compression)
    logger.info(f"Written to {out_csv}")
    logger.info(f"Written to {out_nc}")


def writeCSV(outfile, Lx, csv_order):
    """Write data product to CSV file

    Parameters
    ----------
    outfile : str
        Output file path
    Lx : xr.Dataset
        Dataset to write to file
    csv_order : list
        List order of variables
    """
    Lcsv = Lx.to_dataframe().dropna(how="all")
    if csv_order is not None:
        names = [c for c in csv_order if c in list(Lcsv.columns)]
        Lcsv = Lcsv[names]
    Lcsv.to_csv(outfile)


def writeNC(outfile, Lx, col_names=None, compression=False):
    """Write data product to NetCDF file with compression

    Parameters
    ----------
    outfile : str
        Output file path
    Lx : xr.Dataset
        Dataset to write to file
    """
    if os.path.isfile(outfile):
        os.remove(outfile)
    if col_names is not None:
        names = [c for c in col_names if c in list(Lx.keys())]
    else:
        names = list(Lx.keys())

    encoding = {var: dict() for var in names}

    if compression:
        comp = dict(zlib=True, complevel=4)
        for var in names:
            encoding[var].update(comp)

    Lx[names].to_netcdf(outfile, mode="w", format="NETCDF4", compute=True, encoding=encoding)


def getColNames(vars_df, ds, remove_nan_fields=False):
    """
     Get variable names for a given dataset with respect to its type and processing level

     The dataset must have the the following attributes:
     * level
     * number_of_booms when the processing level is <= 2

     This is mainly for exporting purposes.

    Parameters
     -------
     list
         Variable names
    """
    # selecting variable list based on level
    vars_df = vars_df.loc[vars_df[ds.attrs["level"]] == 1]

    # selecting variable list based on geometry
    if ds.attrs["level"] in ["L0", "L1", "L2"]:
        if ds.attrs["number_of_booms"] == 1:
            vars_df = vars_df.loc[vars_df["station_type"].isin(["one-boom", "all"])]
        elif ds.attrs["number_of_booms"] == 2:
            vars_df = vars_df.loc[vars_df["station_type"].isin(["two-boom", "all"])]

    var_list = list(vars_df.index)
    if remove_nan_fields:
        for v in var_list:
            if v not in ds.keys():
                var_list.remove(v)
                continue
            if ds[v].isnull().all():
                var_list.remove(v)
    return var_list


def addVars(ds, variables):
    """Add variable attributes from file to dataset

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to add variable attributes to
    variables : pandas.DataFrame
        Variables lookup table file

    Returns
    -------
    ds : xarray.Dataset
        Dataset with metadata
    """
    for k in ds.keys():
        if k not in variables.index:
            continue
        ds[k].attrs["standard_name"] = variables.loc[k]["standard_name"]
        ds[k].attrs["long_name"] = variables.loc[k]["long_name"]
        ds[k].attrs["units"] = variables.loc[k]["units"]
        ds[k].attrs["coverage_content_type"] = variables.loc[k]["coverage_content_type"]
        ds[k].attrs["coordinates"] = variables.loc[k]["coordinates"]
    return ds


def addMeta(ds, meta):
    """Add metadata attributes from file to dataset

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to add metadata attributes to
    meta : dict
        Metadata file

    Returns
    -------
    ds : xarray.Dataset
        Dataset with metadata
    """

    # a static latitude, longitude and altitude is saved as attribute along its origin
    var_alias = {"lat": "latitude", "lon": "longitude", "alt": "altitude"}
    for v in ["lat", "lon", "alt"]:
        # saving the reference latitude/longitude/altitude
        original_value = np.nan
        if var_alias[v] in ds.attrs.keys():
            original_value = ds.attrs[var_alias[v]]
        if v in ds.keys():
            # if possible, replacing it with average coordinates based on the extra/interpolated coords
            ds.attrs[var_alias[v]] = ds[v].mean().item()
            ds.attrs[var_alias[v] + "_origin"] = (
                "average of gap-filled postprocessed " + v
            )
        elif "gps_" + v in ds.keys():
            # if possible, replacing it with average coordinates based on the measured coords (can be gappy)
            ds.attrs[var_alias[v]] = ds["gps_" + v].mean().item()
            ds.attrs[var_alias[v] + "_origin"] = (
                "average of GPS-measured " + v + ", potentially including gaps"
            )

        if np.isnan(ds.attrs[var_alias[v]]):
            # if no better data was available to update the coordinate, then we
            # re-use the original value
            ds.attrs[var_alias[v]] = original_value
            ds.attrs[var_alias[v] + "_origin"] = "reference value, origin unknown"

    # Attribute convention for data discovery
    # https://wiki.esipfed.org/Attribute_Convention_for_Data_Discovery_1-3

    # Determine the temporal resolution
    sample_rate = "unknown_sample_rate"
    if len(ds["time"]) > 1:
        time_diff = pd.Timedelta((ds["time"][1] - ds["time"][0]).values)
        if time_diff == pd.Timedelta("10min"):
            sample_rate = "10min"
        elif time_diff == pd.Timedelta("1h"):
            sample_rate = "hourly"
        elif time_diff == pd.Timedelta("1D"):
            sample_rate = "daily"
        elif 28 <= time_diff.days <= 31:
            sample_rate = "monthly"

    if "station_id" in ds.attrs.keys():
        id_components = [
            "dk",
            "geus",
            "promice",
            "station",
            ds.attrs["station_id"],
            ds.attrs["level"],
            sample_rate,
        ]
        ds.attrs["id"] = ".".join(id_components)
    else:
        id_components = [
            "dk",
            "geus",
            "promice",
            "site",
            ds.attrs["site_id"],
            ds.attrs["level"],
            sample_rate,
        ]
        ds.attrs["id"] = ".".join(id_components)

    ds.attrs["history"] = "Generated on " + datetime.datetime.utcnow().isoformat()
    ds.attrs["date_created"] = str(datetime.datetime.now().isoformat())
    ds.attrs["date_modified"] = ds.attrs["date_created"]
    ds.attrs["date_issued"] = ds.attrs["date_created"]
    ds.attrs["date_metadata_modified"] = ds.attrs["date_created"]
    ds.attrs["processing_level"] = ds.attrs["level"].replace("L", "Level ")

    id = ds.attrs.get('station_id', ds.attrs.get('site_id'))
    title_string_format = "AWS measurements from {id} processed to {processing_level}. {sample_rate} average."
    ds.attrs["title"] = title_string_format.format(
        id=id,
        processing_level=ds.attrs["processing_level"].lower(),
        sample_rate=sample_rate.capitalize(),
    )

    if "lat" in ds.keys():
        lat_min = ds["lat"].min().values
        lat_max = ds["lat"].max().values
    elif "gps_lat" in ds.keys():
        lat_min = ds["gps_lat"].min().values
        lat_max = ds["gps_lat"].max().values
    elif "latitude" in ds.attrs.keys():
        lat_min = ds.attrs["latitude"]
        lat_max = ds.attrs["latitude"]
    else:
        lat_min = np.nan
        lat_max = np.nan

    if "lon" in ds.keys():
        lon_min = ds["lon"].min().values
        lon_max = ds["lon"].max().values
    elif "gps_lon" in ds.keys():
        lon_min = ds["gps_lon"].min().values
        lon_max = ds["gps_lon"].max().values
    elif "longitude" in ds.attrs.keys():
        lon_min = ds.attrs["longitude"]
        lon_max = ds.attrs["longitude"]
    else:
        lon_min = np.nan
        lon_max = np.nan

    if "alt" in ds.keys():
        alt_min = ds["alt"].min().values
        alt_max = ds["alt"].max().values
    elif "gps_alt" in ds.keys():
        alt_min = ds["gps_alt"].min().values
        alt_max = ds["gps_alt"].max().values
    elif "altitude" in ds.attrs.keys():
        alt_min = ds.attrs["altitude"]
        alt_max = ds.attrs["altitude"]
    else:
        alt_min = np.nan
        alt_max = np.nan

    ds.attrs["geospatial_bounds"] = (
        "POLYGON(("
        + f"{lat_min} {lon_min}, "
        + f"{lat_min} {lon_max}, "
        + f"{lat_max} {lon_max}, "
        + f"{lat_max} {lon_min}, "
        + f"{lat_min} {lon_min}))"
    )

    ds.attrs["geospatial_lat_min"] = str(lat_min)
    ds.attrs["geospatial_lat_max"] = str(lat_max)
    ds.attrs["geospatial_lon_min"] = str(lon_min)
    ds.attrs["geospatial_lon_max"] = str(lon_max)
    ds.attrs["geospatial_vertical_min"] = str(alt_min)
    ds.attrs["geospatial_vertical_max"] = str(alt_max)

    ds.attrs["geospatial_vertical_positive"] = "up"
    ds.attrs["time_coverage_start"] = str(ds["time"][0].values)
    ds.attrs["time_coverage_end"] = str(ds["time"][-1].values)

    # https://www.digi.com/resources/documentation/digidocs/90001437-13/reference/r_iso_8601_duration_format.htm
    try:
        ds.attrs["time_coverage_duration"] = str(
            pd.Timedelta((ds["time"][-1] - ds["time"][0]).values).isoformat()
        )
        ds.attrs["time_coverage_resolution"] = str(
            pd.Timedelta((ds["time"][1] - ds["time"][0]).values).isoformat()
        )
    except:
        ds.attrs["time_coverage_duration"] = str(pd.Timedelta(0).isoformat())
        ds.attrs["time_coverage_resolution"] = str(pd.Timedelta(0).isoformat())

    # Note: int64 dtype (long int) is incompatible with OPeNDAP access via THREDDS for NetCDF files
    # See https://stackoverflow.com/questions/48895227/output-int32-time-dimension-in-netcdf-using-xarray
    ds.time.encoding["dtype"] = "i4"  # 32-bit signed integer
    # ds.time.encoding["calendar"] = 'proleptic_gregorian' # this is default

    # Load metadata attributes and add to Dataset
    [_addAttr(ds, key, value) for key, value in meta.items()]

    # Check attribute formating
    for k, v in ds.attrs.items():
        if not isinstance(v, str) or not isinstance(v, int):
            ds.attrs[k] = str(v)
    return ds


def _addAttr(ds, key, value):
    """Add attribute to xarray dataset

    ds : xr.Dataset
        Dataset to add attribute to
    key : str
        Attribute name, with "." denoting variable attributes
    value : str/int
        Value for attribute"""
    if len(key.split(".")) == 2:
        try:
            ds[key.split(".")[0]].attrs[key.split(".")[1]] = str(value)
        except:
            pass
            # logger.info(f'Unable to add metadata to {key.split(".")[0]}')
    else:
        ds.attrs[key] = value


def roundValues(ds, df, col="max_decimals"):
    """Round all variable values in data array based on pre-defined rounding
    value in variables look-up table DataFrame

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to round values in
    df : pd.Dataframe
        Variable look-up table with rounding values
    col : str
        Column in variable look-up table that contains rounding values. The
        default is "max_decimals"
    """
    df = df[col]
    df = df.dropna(how="all")
    for var in df.index:
        if var not in list(ds.variables):
            continue
        if df[var] is not np.nan:
            ds[var] = ds[var].round(decimals=int(df[var]))
    return ds


def reformat_time(dataset):
    """Re-format time"""
    t = dataset["time"].values
    dataset["time"] = list(t)
    return dataset


def reformat_lon(dataset, exempt=["UWN", "Roof_GEUS", "Roof_PROMICE", "ORO"]):
    """Switch gps_lon to negative values (degrees_east). We do this here, and
    NOT in addMeta, otherwise we switch back to positive when calling getMeta
    in joinL2"""
    if "station_id" in dataset.attrs.keys():
        id = dataset.attrs["station_id"]
    else:
        id = dataset.attrs["site_id"]

    if id not in exempt:
        if "gps_lon" not in dataset.keys():
            return dataset
        dataset["gps_lon"] = np.abs(dataset["gps_lon"]) * -1
        if "lon" not in dataset.keys():
            return dataset
        dataset["lon"] = np.abs(dataset["lon"]) * -1
    return dataset



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AWS Level 0 (L0) data transmission fetching module
"""

from collections import deque
import email, re, os, os.path, time, unittest, calendar, imaplib, pkg_resources

# Set maximum number of email lines to read
imaplib._MAXLINE = 5000000
    
#------------------------------------------------------------------------------

class PayloadFormat(object):
    '''Payload formatter object'''
    
    def __init__(self, format_file=None, type_file=None):
        '''Payload formatter initialisation
        
        Parameters
        ----------
        format_file : str
            File path to payload formatter .csv file
        type_file : str
            File path to payload type .csv file
        '''
        self.payload_type = self.readType(type_file)
        self.payload_format = self.readFormatter(format_file)
        self._addCount()

    def readType(self, in_file, delimiter=','):                                #TODO read from .txt file as well as .csv
        '''Read payload type setter from file. Outputted dictionary set as
        key[type_letter]: number_of_bytes
        
        Parameters
        ----------
        in_file : str
            Input file path
        delimiter : str, optional
            File delimiter. The default is ","
        
        Returns
        -------
        payload_typ : dict
            Payload type information
        '''
        payload_typ = {}
        if in_file is None:
            lines = self.readPkgFile('tx/payload_types.csv')
        else:
            lines = self.readFile(in_file)
        for l in lines[1:]:
            info = l.split(delimiter)
            try:
                payload_typ[info[0]] = int(info[1])
            except IndexError:
                pass
        return payload_typ
    
    def readFormatter(self, in_file, delimiter=','):
        '''Read payload formatter from file. Outputted dictionary set as
        key[number]: [expected_length, format_characters, description].
        Flag column (info[4]) used to signify if entry should be written to 
        output
        
        Parameters
        ----------
        in_file : str
            Input file path
        delimiter : str, optional
            File delimiter. The default is ","
        
        Returns
        -------
        payload_fmt : dict
            Payload format information
        '''
        payload_fmt = {}
        if in_file==None:
            lines = self.readPkgFile('tx/payload_formats.csv')
        else:
            lines = self.readFile(in_file)
        for l in lines[1:]:
            info = l.split(delimiter)  
            try:                     
                if info[4]:
                    pass
                else:
                    payload_fmt[int(info[0])] = [int(info[1]), info[2], info[3]]
            except IndexError:
                pass
        return payload_fmt                 

    def readFile(self, in_file):
        '''Read lines from file
        
        Parameters
        ----------
        in_file : str
            Input file path
        
        Returns
        -------
        lines : list
            List of file line contents
        '''
        with open(in_file, 'r') as in_f:
            lines = in_f.readlines()
        return lines

    def readPkgFile(self, fname):
        '''Read lines from internal package file
        
        Parameters
        ----------
        fname : str
            Package file name
        
        Returns
        -------
        lines : list
            List of file line contents
        '''
        with pkg_resources.resource_stream('pypromice', fname) as stream:
            lines = stream.read().decode("utf-8")
            lines = lines.split("\n")  
        return lines
 
    def _addCount(self):
        '''Add counter to payload formatter'''
        for item in self.payload_format.items():
            key, val = item
            var_count, var_def, comment = val
            assert var_count == len(var_def)
            bytes_count = 0
            for var in var_def:
                bytes_count += self.payload_type[var.lower()]
            self.payload_format[key].append(bytes_count + 1)
#------------------------------------------------------------------------------

class SbdMessage(object):
    '''SBD transmission message object'''

    def __init__(self, content, attach, imei):
        '''SBD tranmission message initialisation
        
        Parameters
        ----------
        content : str
            Email message string
        attach : email.message.Message
            Attachment object
        imei : str
            Modem number of message
        '''          
        # Get email information
        self.imei = imei
        if content:
            self.momsn = self.getKeyValue(content, ': ', 'MOMSN')
            self.mtmsn = self.getKeyValue(content, ': ', 'MTMSN')
            self.cep_radius = self.getKeyValue(content, '= ', 'CEPradius')
            self.msg_size = self.getKeyValue(content, ': ', 'Message Size (bytes)')  
            self.session_time = self.getKeyValue(content, ': ', 
                                                 'Time of Session (UTC)', 
                                                 integer=False)
            self.status = self.getStatus(content)
            self.location = self.getLocation(content)
        
        # Create empty object if no email content
        else:
            self.momsn = None
            self.mtmsn = None
            self.cep_radius = None
            self.msg_size = None
            self.session_time = None
            self.status = None
            self.location = None
        
        # Get attachment information                                           #TODO read from downloaded file as alternative to attachment                  
        if attach:
            if self.checkAttachment(attach) and self.checkAttachmentName(attach.get_filename()): 
                self.payload = self.getPayloadFromEmail(attach, self.msg_size)
            else:
                self.payload = None
        else:
            self.payload = None

    def checkAttachment(self, attach):
        '''Check if attachment is present in email.message.Message object'''
        if attach == 0:
            print('No file attached to email')
            return False
        else:
            return True

    def checkAttachmentName(self, attach_file):
        '''Check if attachment is .sbd file'''
        root, ext = os.path.splitext(attach_file)
        if ext == '.sbd' or ext == '.dat':
            return True
        else:
            print(f'Unrecognised attachment file type: {attach_file}')
            return False
    
    def getPayloadFromEmail(self, attach, message_size):
        '''Get Sbd payload from email object'''
        if attach != None:
            sbd_payload = attach.get_payload(decode=True)
            if message_size:
                assert len(sbd_payload) == message_size                        #TODO Provide if statement in case message_size not provided
        else:
            sbd_payload = None
        return sbd_payload                    
    
    def getPayloadFromFile(self, attach):
        '''Read Sbd payload from .sbd file'''
        return readSBD(attach)
             
    def getKeyValue(self, content, seps, key, integer=True):
        '''Get attribute from email via keyword'''
        line = findLine(content, key)
        value = parseValue(line, seps)
        if integer:
            try:
                value = int(value)
            except:
                print(f'Value in line could not be passed as integer: "{line}"')
                value = None
        return value
    
    def getStatus(self, content, seps1=': ', seps2=' ', key='Session Status'):
        '''Get session status from email message'''
        line = findLine(content, key)
        try:
            value = parseValue(line, seps1).split(seps2)
            return [int(value[0]), value[2]]
        except:
            print(f'Session status not parsed from line "{line}"')
            return None
                
    def getLocation(self, content, seps=' ', key='Unit Location'):
        '''Get latitude longitude unit location from email message'''
        line = findLine(content, key)
        try:
            l = line.split(seps)
            if l[2].lower() == 'lat':
                lat = float(l[4])
            
            if l[5].lower() == 'long' or l[5].lower() == 'lon':
                lon = float(l[7])
            return [lat, lon]
        except:
            print(f'Unit location not parsed from line "{line}"')
            return None

#------------------------------------------------------------------------------

class EmailMessage(SbdMessage):                                                #TODO add initialisation from .sbd file  
    '''Email message object'''

    def __init__(self, email_msg, sender_name): 
        '''Email message initialisation
        
        Parameters
        ----------
        email_msg : email.message.Message
            Email message object
        sender_name : str or list
            Email sender name/s to filter by. Filtering by sender name does not
            occur if variable is None
        '''
        if self.checkEmail(email_msg):
            
            self._email_msg = email_msg
            
            # Parse email data
            self.email_data = {}
            self.email_data = self.getEmailInfo()                              #TODO Get metadata

            # Check sender 
            if self.checkSender(sender_name):
                content, attachment = self.getEmailBody()  
                SbdMessage.__init__(self, content, attachment, self.getIMEI())
            else:
                print('Email from unrecognised sender ' + 
                      f'"f{self.email_data["from"]}", no Sbd message read')
                SbdMessage.__init__(self, None, None, None)
            
        else:
            print('Empty email message')
            self._email_msg = None
            self.email_data = None
        
    def checkEmail(self, email_msg):
        '''Check if email is Message object'''
        if not isinstance(email_msg, email.message.Message):
            print(f'Email object not valid {type(email_msg)}. ' +
                  'Expecting email.message.Message object')
            return False
        else:
            return True

    def checkSender(self, sender_name):
        '''Check email message from field matches sender name or names'''
        sender = self.email_data['from']
        if isinstance(sender_name, str):
            flag = self._checkFrom(sender, sender_name)
        elif isinstance(sender_name, list):
            flag = False
            for s in sender_name:
                f = self._checkFrom(sender, s)
                if f==True: 
                    flag = True
        else:
            flag = True
        return flag

    def getEmailBody(self):
        '''Get email message body'''
        if self._email_msg.is_multipart():
            try:
                content, attachment = self._email_msg.get_payload()  
            except:
                content = self._email_msg.get_payload()[0]
                attachment = None
                
            assert not content.is_multipart()                                  #else the decode=True on the next line makes it 
            body = content.get_payload(decode=False)                           #return None and break the rest of the parsing
        else:
            body = self._email_msg.get_payload(decode=True)#[0]
            attachment=None
        return body, attachment
    
    def getIMEI(self):
        '''Get modem identifier from email subject string'''
        try:
            imei, = re.findall(r'[0-9]+', self.email_data['subject'])
        except:
            imei=None
        return imei
        
    def getEmailInfo(self):
        '''Parse message in email object'''
        email_data = {}
        for i in ['from', 'subject', 'date']:
            email_data[i] = self._getEmailData(i)[0]
        email_data['to'] = self._getEmailData('to')            
        return email_data
    
    def _getEmailData(self, descriptor):
        '''Get email data from descriptor'''
        return self._email_msg.get_all(descriptor)
    
    def _checkFrom(self, email_from, name):
        '''Check if name is in email sender string'''
        if name not in email_from:
            return False
        else: 
            return True

#------------------------------------------------------------------------------    

class L0tx(EmailMessage, PayloadFormat):
    '''L0 tranmission data object'''

    def __init__(self, email_msg, format_file=None, type_file=None, 
                 sender_name=['sbdservice', 'ice@geus.dk','emailrelay@konectgds.com'],    #TODO don't hardcode sender names?
                 UnixEpochOffset=calendar.timegm((1970,1,1,0,0,0,0,1,0)),
                 CRbasicEpochOffset = calendar.timegm((1990,1,1,0,0,0,0,1,0))):
        '''L0tx object initialisation. 
        
        A note on the epoch, the Win32 epoch is 1st Jan 1601 but MSC epoch is 
        1st Jan 1970 (MSDN gmtime docs), same as Unix epoch. Neither Python nor 
        ANSI-C explicitly specify any epoch but CPython relies on the 
        underlying C library. CRbasic instead has the SecsSince1990() function. 
        This should always evaluate to 0 in CPython on Win32
        
        Parameters
        ----------
        email_msg : email.message.Message
            Email message object
        format_file : str
            File path to payload formatter .csv file
        type_file : str
            File path to payload type .csv file
        sender_name : str or list
            Valid sender names in email
        UnixEpochOffset : int
            Unix/MSC epoch 
        CRbasicEpochOffset : int
            CRbasic epoch
        '''
        # Initialise EmailMessage and PayloadFormatter objects
        EmailMessage.__init__(self, email_msg, sender_name)
        PayloadFormat.__init__(self, format_file, type_file)
          
        # Set epoch offset
        self.EpochOffset = UnixEpochOffset + CRbasicEpochOffset

        # Process if payload is present and in a compliant format
        if self.checkPayload() and self.checkByte(self.payload[:1]):
            
            self.bin_val, self.bin_format, self.bin_name, self.bin_len, self.bin_idx, self.bin_valid = self.getFormat()
            msg = self.getDataLine() 
            if msg:
                if self.bin_valid: 
                    
                    if self.checkLength():
                        self.flag = '-F'
                    elif self.isDiagnostics(msg):
                        self.flag = '-D'
                    elif self.isObservations(msg):
                        self.flag = ''
                    else:
                        self.flag = '-X'
                else:
                    self.flag=''
                
                self.msg = msg
            
            else:
                self.msg = None
        else:
            self.msg = None
    
    def checkPayload(self):
        '''Check message payload'''
        if self.payload == None:
           print('No .sbd file attached to this message')
           return False
        else:
            return True

    def getFirstByte(self):
        '''Get first byte in payload'''
        return self.payload[:1]
        
    def getFormat(self):
        '''Get binary format type from first byte in payload
        
        Returns
        -------
        bval : int or None
            Format value
        bfor : str or None
            Format string characters
        bname : str or None
            Format name
        blength : int or None
            Expected format length
        bidx : int
            Format index
        bool
            Valid format flag
        '''
        if self.getFirstByte().isdigit() or (self.payload[:2] == '\n' and self.imei == 300234064121930):     #TODO needed?
            return None, None, None, None, -9999, False
        
        elif 'watson' in self.email_data['subject'].lower() or 'gios' in self.email_data['subject'].lower():
            return None, None, None, None, -9999, False    
       
        else:
            bidx = ord(self.getFirstByte()) 
            try:
                bval, bfor, bname, blength = self.payload_format[bidx]
                
                # if bidx==80:                                                   # TODO. This is a temporary workaround for QAS_Lv3 formatting being different. Needs a more permanent fix!
                #     print('Checking version 3...')
                #     if len(self.payload) != blength:
                #         print('Mismatched lengths found...')
                #         if len(self.payload) == 83:
                #             print('Fetching QAS_Lv3 specific formatting...')
                #             bval, bfor, bname, blength = self.payload_format[70]
                
                return bval, bfor, bname, blength, bidx, True
            except KeyError:
                return None, None, None, None, bidx, False
            
    def checkByte(self, b):
        '''Check byte format against payload formatter object'''
        if ord(b) not in self.payload_format:
            print('Unrecognized first byte %s' %hex(ord(b)))
            return False      
        else:    
            return True
    
    def checkLength(self):                                                   
        if len(self.payload) != self.bin_len:
            print(f'Message malformed: expected {self.bin_len} bytes, ' + 
                  f'found {len(self.payload)}. Missing values to be replaced by '+
                  '"?" and extra values dropped')
            return True
        else:
            return False
 
    def check2BitNAN(self, msg, type_letter, letter_flag=['g','n','e'], nan_value=8191):
        '''Check if byte is a 2-bit NAN. This occurs when the GPS data is not 
        available and the logger sends a 2-bytes NAN instead of a 4-bytes value
        '''
        if type_letter.lower() in letter_flag:
            ValueBytes = self.getByteValue(2, msg, self.bytecounter)
            try:
                if GFP2toDEC(ValueBytes) == nan_value:
                    return True
                else:                                
                    return False
            except:
                return False
        else:
            return False
  
    def isDiagnostics(self, DataLine):
        '''Flag if message is diagnostics'''
        return '!D' in DataLine[-5:-3] or self.bin_idx % 5 == 4             

    def isObservations(self, DataLine):
        '''Flag if message is observations''' 
        return '!M' in DataLine[-2:] or self.bin_valid
    
    def isSummer(self, DataLine):
        '''Flag if message is summer message'''
        return ('!S' in DataLine and '!M' in DataLine[-2:]) or self.bin_idx % 5 in (0, 1)
    
    def isWatsonObservation(self, DataLine):
        '''Flag if message is Watson River measurement'''
        return ('watson' in DataLine.lower() or 'gios' in DataLine.lower())
    
    def isWithInstance(self, DataLine):
        '''Flag if message is with instance'''
        return '!I' in DataLine[-5:-3] or (self.bin_idx % 5 in (1, 3) and self.bin_idx != -9999)

    def getByteValue(self, ValueBytesCount, BinaryMessage, idx):
        '''Get values from byte range in binary message'''
        ValueBytes=[]
        for i in range(0,ValueBytesCount):
            try:
                ValueBytes.append(ord(BinaryMessage[idx+i: idx+i+1]))
            except:
                print('No byte found')
        return ValueBytes
    
    def updateByteCounter(self, value):
        '''Update byte counter for decoding message'''
        self.bytecounter += value

    def writeEntry(self, entry, i):
        '''Write out comma-formatted data entry from message'''
        if i == self.bin_val-1:
            return entry
        else:
            return entry + ','        
               
    def getDataLine(self):                                                     #TODO clean up
        '''Get data line from transmission message
        
        Returns
        -------
        str or None
            Dataline string if found'''
        # Retrieve payload and prime object if valid binary message        
        bin_msg = self.payload[1:]
        if self.bin_valid and bin_msg:
            print('%s-%s (binary)' %(self.imei, self.momsn) , self.bin_name)

            dataline = ''
            self.bytecounter = 0
            
            # Iterate over binary message formatting string
            for i in range(0, self.bin_val):

                type_letter = self.bin_format[i]
                num_bytes = self.payload_type[type_letter.lower()]
                
                # Check if 2-bit NaN is present
                if self.check2BitNAN(bin_msg, type_letter):
                    dataline = dataline + self.writeEntry('NAN', i)
                    self.updateByteCounter(2)
                    self.bin_len -= 2
                
                # Get byte value
                else:                      
                    ValueBytes = self.getByteValue(num_bytes, 
                                                   bin_msg, self.bytecounter)
                    self.updateByteCounter(num_bytes)
                    
                    if len(ValueBytes) == 2:
                        Value = GFP2toDEC(ValueBytes)
                    elif len(ValueBytes) == 4:
                        Value = GLI4toDEC(ValueBytes)
                    else:
                        entry = '?'
                    
                    # Decode based on formatting string                        #TODO put this in payload_type file
                    if type_letter.lower()=='g':
                        entry = str(Value/100.0)
                        
                    elif type_letter.lower()=='n':
                        entry = str(Value/100000.0)
                        
                    elif type_letter.lower() =='e':
                        entry = str(Value/100000.0)  
                        
                    elif type_letter.lower()=='f':
                        if Value == 8191:
                            entry = 'NAN'
                        elif Value == 8190:
                            entry = 'INF'
                        elif Value == -8190 or Value == -8191:               
                            entry = '-INF'
                        else:
                            entry = str(Value)    
                            
                    elif type_letter.lower()=='l':
                        if Value in (-2147483648, 2147450879):
                            entry = 'NAN'
                        else:
                            entry = str(Value)
                    
                    elif type_letter.lower()=='t':
                        entry = time.strftime('%Y-%m-%d %H:%M:%S', 
                                              time.gmtime(Value + self.EpochOffset)) + ',' + str(Value)  
                    else:
                        entry = '?'
                                        
                    # Append value outputted dataline
                    if type_letter.isupper():                                  #TODO test RAWtoSTR
                        dataline = dataline + RAWtoSTR(ValueBytes)
                    else:
                        dataline = dataline + self.writeEntry(entry, i) 
                    
            return dataline
        
        else:
            try:
                bin_msg = '2' + bin_msg.decode('cp850')                        #TODO de-bug so first byte is passed (currently misses of the first "2" of the year e.g. "022" instead of "2022")
            except:
                bin_msg = ''
            if self.isDiagnostics(bin_msg):
                desc = f'{self.imei}-{self.momsn} ASCII generic diagnostic message'
            elif self.isObservations(bin_msg) and self.isSummer(bin_msg):
                desc = f'{self.imei}-{self.momsn} ASCII generic summer observations message'
            elif self.isObservations(bin_msg) and not self.isSummer(bin_msg):
                desc = f'{self.imei}-{self.momsn} ASCII generic winter observations message'
            elif self.isWatsonObservation(bin_msg):
                desc = 'Watson River observations message'
                bin_msg = bin_msg.split('"Smp"')[-1].replace('"', '')
                    
            else:
                desc=None
                
            if desc:
                if self.isWithInstance(bin_msg):
                    desc = desc + '(+ instant.)'
                print(desc)
                return bin_msg
            else:
                print('Unrecognized message format')
                return None
                   
#------------------------------------------------------------------------------

def GFP2toDEC(Bytes):
    '''Two-bit decoder
    
    Parameters
    ----------
    Bytes : list
        List of two values
    
    Returns
    -------
    float
        Decoded value
    '''
    # print('ValueBytes received ' + str(Bytes))
    msb = Bytes[0]
    lsb = Bytes[1]    
    Csign = -2*(msb & 128)/128 + 1
    CexpM = (msb & 64)/64
    CexpL = (msb & 32)/32
    Cexp = 2*CexpM + CexpL - 3
    Cuppmant = 4096*(msb & 16)/16 + 2048*(msb & 8)/8 + 1024*(msb & 4)/4 + 512*(msb & 2)/2 + 256*(msb & 1)
    Cnum = Csign * (Cuppmant + lsb)*10**Cexp
    return round(Cnum, 3)

def GLI4toDEC(Bytes):
    '''Four-bit decoder
    
    Parameters
    ----------
    Bytes : list
        List of four values
    
    Returns
    -------
    float
        Decoded value
    '''
    Csign = int(-2 * (Bytes[0] & 0x80) / 0x80 + 1)
    byte1 = Bytes[0] & 127
    byte2 = Bytes[1]
    byte3 = Bytes[2]
    byte4 = Bytes[3]    
    return Csign * byte1 * 0x01000000 + byte2 * 0x010000 + byte3 * 0x0100 + byte4

def RAWtoSTR(Bytes):
    '''Byte-to-string decoder
    
    Parameters
    ----------
    Bytes : list
        List of values
    
    Returns
    -------
        Decoded string characters
    '''
    us = [chr(byte) for byte in Bytes] #the unicode strings
    hs = ['0x{0:02X}'.format(byte) for byte in Bytes] #the hex strings
    bs = ['0b{0:08b}'.format(byte) for byte in Bytes] #the bit strings
    return '(%s = %s = %s)' %(' '.join(us), ' '.join(hs), ' '.join(bs))
    
def findLine(content, key):
    '''Find keyword in line
    
    Parameters
    ----------
    content : str
        String to find keyword in
    key : str
        Keyword to find in string
   
    Returns
    -------
    line : str
       Line that keyword appears on
    '''
    for line in content.splitlines():
        if key in line:   
            return line

def parseValue(line, seps):
    '''Parse last value from line according to separating characters
    
    Parameters
    ----------
    line : str
        String to split 
    sep : str
        Separator characters to split line by

    Returns
    -------
    value : str
        Value extracted from line
    '''
    try:
        value = line.split(seps)[-1]
    except:
        print(f'Value not read from line "{line}"')
        value = None
    return value

def getMail(mail_server, last_uid=1):
    '''Retrieve new mail

    Parameters
    ----------
    mail_server : imaplib.IMAP_SSL
        Mail server object
    last_uid : int, optional
        Mail uid to start retrieval from. The default is 1.

    Yields
    ------
    str
        Mail uid
    str
        Mail message
    '''
    # Issue search command of the form "SEARCH UID 42:*"
    command = '(UID {}:*)'.format(last_uid)
    result, data = mail_server.uid('search', None, command)
    messages = data[0].split()
    new_uids = data[0].decode()
    # drop the last_uid (it has already been processed)
    new_uids = new_uids.replace(str(last_uid), '')
    print('Newest UID: %s' % new_uids.split(' ')[-1])
    # print('new UIDs: %s' % new_uids)

    # Yield mails
    for message_uid in messages:
        
        # SEARCH command *always* returns at least the most
        # recent message, even if it has already been synced
        if int(message_uid) > last_uid:
            print(f'\nFetching mail {message_uid.decode()}')
            result, data = mail_server.uid('fetch', message_uid, '(RFC822)')
            
            # Yield raw mail body
            yield message_uid.decode(), data[0][1].decode() 

def loadMsg(fname):
    '''Load .msg email file into format compatible with EmailMessage and 
    SbdMessage objects
    
    Parameters
    ----------
    fname : str
        File path to .msg file
    
    Returns
    -------
    email.message.Message
        Email message object
    '''
    with open(fname, 'rb') as f:
        byte = f.read()
    return email.message_from_bytes(byte)

def saveMsg(msg, fname):
    '''Save email message object to .msg file
    
    Parameters
    ----------
    msg : email.message.Message
        Email object to save to file
    fname : str
        File path to outputted .msg file
    '''
    with open(fname, 'wb') as fp:
        fp.write(bytes(msg))
    
def readSBD(sbd_file):
    '''Read encoded .sbd transmission file
    
    Parameters
    ----------
    sbd_file : str
        Filepath to encoded .sbd file
    
    Returns
    -------
    data : bytes
        Transmission message byte object
    '''
    with open(sbd_file, 'rb') as file:
         data = file.readlines()  
    return data[0]

def findDuplicates(lines):
    '''Find duplicates lines in list of strings
    
    Parameters
    ---------
    lines : list
       List of strings
    
    Returns
    -------
    unique_lines : list
       List of unique strings
    '''
    unique_lines = list(set(lines))
    duplicates_count = len(lines) - len(unique_lines)
    print(f'{duplicates_count} duplicates found')   
    return unique_lines

def sortLines(in_file, out_file, replace_unsorted=True):                       #Formerly called sorter.py
    '''Sort lines in text file

    Parameters
    ----------
    in_file : str
        Input file path
    out_file : str
        Output file path
    replace_unsorted : bool, optional
        Flag to replace unsorted files with sorted files. The default is True.
    '''
    print(f'\nSorting {in_file}')
    
    # Open input file and read lines
    with open(in_file) as in_f:
        lines = in_f.readlines()
    
    # Remove duplicate lines and sort
    unique_lines = findDuplicates(lines.copy())
    unique_lines.sort()
    if lines != unique_lines:
        # Write sorted file
        with open(out_file, 'w') as out_f:
            # out_f.write(headers)
            out_f.writelines(unique_lines)
    
        # Replace input file with new sorted file
        if replace_unsorted:
            os.remove(in_file)
            os.rename(out_file, in_file)

def addTail(in_file, out_dir, aws_name, header_names='', lines_limit=100):    
    '''Generate tails file from L0tx file

    Parameters
    ----------
    in_file : str
        Input L0tx file
    out_dir : str
        Output directory for tails file
    aws_name : str
        AWS name
    header_names : str, optional
        Header names. The default is ''.
    lines_limit : int, optional
        Number of lines to append to tails file. The default is 100.
    '''
    with open(in_file) as in_f:
        tail = deque(in_f, lines_limit)
    
    headers_lines = [l + '\n' for l in header_names.split('\n')]
    tail = list(set(headers_lines + list(tail)))
    tail.sort()
        
    out_fn = '_'.join((aws_name, in_file.split('/')[-1]))
    out_pn = os.sep.join((out_dir, out_fn))
    
    with open(out_pn, 'w') as out_f:
        #out_f.write(headers)
        out_f.writelines(tail) 
        print(f'Tails file written to {out_pn}')

def isModified(filename, time_threshold=1):
    '''Return flag denoting if file is modified within a certain timeframe
    
    Parameters
    ----------
    filename : str
        File path
    time_threshold : int
        Time threshold (provided in hours)
    
    Returns
    -------
    bool
        Flag denoting if modified (True) or not (False)
    '''
    delta = time.time() - os.path.getmtime(filename)
    delta = delta / (60*60)
    if delta < time_threshold:
        return True
    return False

def _loadTestMsg():
    '''Load test .msg email file'''
    with pkg_resources.resource_stream('pypromice', 'test/test_email') as stream:
        byte = stream.read()
    return email.message_from_bytes(byte)

#------------------------------------------------------------------------------
        
class TestTX(unittest.TestCase): 
    def testPayloadFormat(self):
        '''Test PayloadFormat object initialisation'''
        p = PayloadFormat()
        self.assertTrue(p.payload_format[30][0]==12)
    
    def testEmailMessage(self):   
        '''Test EmailMessage object initialisation from .msg file'''
        m = _loadTestMsg()
        e = EmailMessage(m, sender_name='sbdservice')
        self.assertEqual(e.momsn, 36820)
        self.assertEqual(e.msg_size, 27)
        self.assertTrue(e.imei in '300234061165160')
        self.assertFalse(e.mtmsn)
        
    def testL0tx(self):
        '''Test L0tx object initialisation'''
        m = _loadTestMsg()
        l0 = L0tx(m)
        self.assertTrue(l0.bin_valid)
        self.assertEqual(l0.bin_val, 12)
        self.assertTrue('tfffffffffff' in l0.bin_format)
        self.assertTrue('2022-07-25 10:00:00' in l0.msg)
        self.assertFalse('?' in l0.msg)

    def testCLIl0tx(self):
        '''Test get_l0tx CLI'''
        exit_status = os.system('get_l0tx -h')
        self.assertEqual(exit_status, 0)
 
    def testCLIwatson(self):
        '''Test get_watsontx CLI'''
        exit_status = os.system('get_watsontx -h')
        self.assertEqual(exit_status, 0)
 
    def testCLImsg(self):
        '''Test get_msg CLI'''
        exit_status = os.system('get_msg -h')
        self.assertEqual(exit_status, 0)
                   
if __name__ == "__main__":  
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 10 10:58:39 2024

@author: pho
"""
import logging
import numpy as np
import xarray as xr
from pypromice.process.L1toL2 import calcDirWindSpeeds
logger = logging.getLogger(__name__)

def resample_dataset(ds_h, t):
    '''Resample L2 AWS data, e.g. hourly to daily average. This uses pandas
    DataFrame resampling at the moment as a work-around to the xarray Dataset
    resampling. As stated, xarray resampling is a lengthy process that takes
    ~2-3 minutes per operation: ds_d = ds_h.resample({'time':"1D"}).mean()
    This has now been fixed, so needs implementing:
    https://github.com/pydata/xarray/issues/4498#event-6610799698

    Parameters
    ----------
    ds_h : xarray.Dataset
        L3 AWS dataset either at 10 min (for raw data) or hourly (for tx data)
    t : str
        Resample factor, same variable definition as in
        pandas.DataFrame.resample()

    Returns
    -------
    ds_d : xarray.Dataset
        L3 AWS dataset resampled to the frequency defined by t
    '''
    # Convert dataset to DataFrame
    df_d = ds_h.to_dataframe()
    
    # Identify non-numeric columns
    non_numeric_cols = df_d.select_dtypes(exclude=['number']).columns
    
    # Log a warning and drop non-numeric columns
    if len(non_numeric_cols) > 0:
        for col in non_numeric_cols:
            unique_values = df_d[col].unique()
            logger.warning(f"Dropping column '{col}' because it is of type '{df_d[col].dtype}' and contains unique values: {unique_values}")

        df_d = df_d.drop(columns=non_numeric_cols)
    # Resample the DataFrame
    df_d = df_d.resample(t).mean()

    # taking the 10 min data and using it as instantaneous values:
    is_10_minutes_timestamp = (ds_h.time.diff(dim='time') / np.timedelta64(1, 's') == 600)
    if (t == '60min') and is_10_minutes_timestamp.any():
        cols_to_update = ['p_i', 't_i', 'rh_i', 'rh_i_wrt_ice_or_water', 'wspd_i', 'wdir_i','wspd_x_i','wspd_y_i']
        cols_origin = ['p_u', 't_u', 'rh_u', 'rh_u_wrt_ice_or_water', 'wspd_u', 'wdir_u','wspd_x_u','wspd_y_u']
        timestamp_10min = ds_h.time.where(is_10_minutes_timestamp, drop=True).to_index()
        timestamp_round_hour = df_d.index
        timestamp_to_update = timestamp_round_hour.intersection(timestamp_10min)
        
        for col, col_org in zip(cols_to_update, cols_origin):
            if col not in df_d.columns:
                df_d[col] = np.nan
            else:
                # if there are already instantaneous values in the dataset
                # we want to keep them as they are
                # removing timestamps where there is already t_i filled from a TX file
                missing_instantaneous = ds_h.reindex(time=timestamp_to_update)[col].isnull()
                timestamp_to_update = timestamp_to_update[missing_instantaneous]
            df_d.loc[timestamp_to_update, col] = ds_h.reindex(
                time= timestamp_to_update
                )[col_org].values
            if col == 'p_i':
                df_d.loc[timestamp_to_update, col] = df_d.loc[timestamp_to_update, col].values-1000
            

    # recalculating wind direction from averaged directional wind speeds
    for var in ['wdir_u','wdir_l']:
        boom = var.split('_')[1]
        if var in df_d.columns:
            if ('wspd_x_'+boom in df_d.columns) & ('wspd_y_'+boom in df_d.columns):
                df_d[var] = _calcWindDir(df_d['wspd_x_'+boom], df_d['wspd_y_'+boom])
            else:
                logger.info(var+' in dataframe but not wspd_x_'+boom+' nor wspd_y_'+boom+', recalculating them')
                ds_h['wspd_x_'+boom], ds_h['wspd_y_'+boom] = calcDirWindSpeeds(ds_h['wspd_'+boom], ds_h['wdir_'+boom])
                df_d[['wspd_x_'+boom, 'wspd_y_'+boom]] = ds_h[['wspd_x_'+boom, 'wspd_y_'+boom]].to_dataframe().resample(t).mean()
                df_d[var] = _calcWindDir(df_d['wspd_x_'+boom], df_d['wspd_y_'+boom])
    
    # recalculating relative humidity from average vapour pressure and average
    # saturation vapor pressure
    for var in ['rh_u','rh_l']:
        lvl = var.split('_')[1]
        if var in df_d.columns:
            if ('t_'+lvl in ds_h.keys()):
                es_wtr, es_cor = calculateSaturationVaporPressure(ds_h['t_'+lvl])
                p_vap = ds_h[var] / 100 * es_wtr
                
                df_d[var] = (p_vap.to_series().resample(t).mean() \
                           / es_wtr.to_series().resample(t).mean())*100
                if var+'_wrt_ice_or_water' in df_d.keys():
                    df_d[var+'_wrt_ice_or_water'] = (p_vap.to_series().resample(t).mean() \
                               / es_cor.to_series().resample(t).mean())*100
    
    # passing each variable attribute to the ressample dataset
    vals = []
    for c in df_d.columns:
        if c in ds_h.data_vars:
            vals.append(xr.DataArray(
                data=df_d[c], dims=['time'],
               coords={'time':df_d.index}, attrs=ds_h[c].attrs))
        else:
            vals.append(xr.DataArray(
                data=df_d[c], dims=['time'],
               coords={'time':df_d.index}, attrs=None))
            
    ds_d = xr.Dataset(dict(zip(df_d.columns,vals)), attrs=ds_h.attrs)
    return ds_d


def calculateSaturationVaporPressure(t, T_0=273.15, T_100=373.15, es_0=6.1071,
                                     es_100=1013.246, eps=0.622):            
    '''Calculate specific humidity
    
    Parameters
    ----------
    T_0 : float 
        Steam point temperature. Default is 273.15.
    T_100 : float
        Steam point temperature in Kelvin
    t : xarray.DataArray
        Air temperature
    es_0 : float
        Saturation vapour pressure at the melting point (hPa)
    es_100 : float
        Saturation vapour pressure at steam point temperature (hPa)
    
    Returns
    -------
    xarray.DataArray
        Saturation vapour pressure with regard to water above 0 C (hPa)
    xarray.DataArray
        Saturation vapour pressure where subfreezing timestamps are with regards to ice (hPa)
    '''                                                         
    # Saturation vapour pressure above 0 C (hPa)
    es_wtr = 10**(-7.90298 * (T_100 / (t + T_0) - 1) + 5.02808 * np.log10(T_100 / (t + T_0))
                  - 1.3816E-7 * (10**(11.344 * (1 - (t + T_0) / T_100)) - 1)
                  + 8.1328E-3 * (10**(-3.49149 * (T_100 / (t + T_0) -1)) - 1) + np.log10(es_100))

    # Saturation vapour pressure below 0 C (hPa)
    es_ice = 10**(-9.09718 * (T_0 / (t + T_0) - 1) - 3.56654
                  * np.log10(T_0 / (t + T_0)) + 0.876793
                  * (1 - (t + T_0) / T_0)
                  + np.log10(es_0)) 
    
    # Saturation vapour pressure (hPa)
    es_cor = xr.where(t < 0, es_ice, es_wtr)
    
    return es_wtr, es_cor

def _calcWindDir(wspd_x, wspd_y):
    '''Calculate wind direction in degrees

    Parameters
    ----------
    wspd_x : xarray.DataArray
        Wind speed in X direction
    wspd_y : xarray.DataArray
        Wind speed in Y direction

    Returns
    -------
    wdir : xarray.DataArray
        Wind direction'''
    deg2rad = np.pi / 180
    rad2deg = 1 / deg2rad
    wdir = np.arctan2(wspd_x, wspd_y) * rad2deg
    wdir = (wdir + 360) % 360
    return wdir



#!/usr/bin/env python
import logging, sys, toml
from pathlib import Path

import xarray as xr
from argparse import ArgumentParser
import pypromice
from pypromice.process.L2toL3 import toL3
import pypromice.resources
from pypromice.process.write import prepare_and_write
logger = logging.getLogger(__name__)

def parse_arguments_l2tol3(debug_args=None):
    parser = ArgumentParser(description="AWS L3 script for the processing L3 "+
                            "data from L2. An hourly, daily and monthly L3 "+
                            "data product is outputted to the defined output path")
    parser.add_argument('-c', '--config_folder', type=str, required=True,
                        default='../aws-l0/metadata/station_configurations/',
                        help='Path to folder with sites configuration (TOML) files')
    parser.add_argument('-i', '--inpath', type=str, required=True, 
                        help='Path to Level 2 .nc data file')
    parser.add_argument('-o', '--outpath', default=None, type=str, required=False, 
                        help='Path where to write output')
    parser.add_argument('-v', '--variables', default=None, type=str, 
                        required=False, help='File path to variables look-up table')
    parser.add_argument('-m', '--metadata', default=None, type=str, 
                        required=False, help='File path to metadata')
    parser.add_argument('--data_issues_path', '--issues', default=None, help="Path to data issues repository")


    args = parser.parse_args(args=debug_args)
    return args

def get_l2tol3(config_folder: Path|str, inpath, outpath, variables, metadata, data_issues_path: Path|str):
    if isinstance(config_folder, str):
        config_folder = Path(config_folder)

    logging.basicConfig(
        format="%(asctime)s; %(levelname)s; %(name)s; %(message)s",
        level=logging.INFO,
        stream=sys.stdout,
    )
  
    # Define Level 2 dataset from file
    with xr.open_dataset(inpath) as l2:
        l2.load()
    
    # Remove encoding attributes from NetCDF
    for varname in l2.variables:
        if l2[varname].encoding!={}:
            l2[varname].encoding = {}

    if 'bedrock' in l2.attrs.keys():
        l2.attrs['bedrock'] = l2.attrs['bedrock'] == 'True'
    if 'number_of_booms' in l2.attrs.keys():
        l2.attrs['number_of_booms'] = int(l2.attrs['number_of_booms'])
    
    # importing station_config (dict) from config_folder (str path)
    config_file = config_folder / (l2.attrs['station_id']+'.toml')

    if config_file.exists():
        # File exists, load the configuration
        station_config = toml.load(config_file)
    else:
        # File does not exist, initialize with standard info 
        # this was prefered by RSF over exiting with error
        logger.error("\n***\nNo station_configuration file for %s.\nPlease create one on AWS-L0/metadata/station_configurations.\n***"%l2.attrs['station_id'])
        station_config = {"stid":l2.attrs['station_id'],
                        "station_site":l2.attrs['station_id'],
                        "project": "PROMICE",
                        "location_type": "ice sheet",
                        }
        
    # checking that the adjustement directory is properly given
    if data_issues_path is None:
        data_issues_path = Path("../PROMICE-AWS-data-issues")
        if data_issues_path.exists():
            logging.warning(f"data_issues_path is missing. Using default data issues path: {data_issues_path}")
        else:
            raise ValueError("data_issues_path is missing. Please provide a valid path to the data issues repository")
    else:
        data_issues_path = Path(data_issues_path)

    data_adjustments_dir = data_issues_path / "adjustments"
    
    # Perform Level 3 processing
    l3 = toL3(l2, data_adjustments_dir, station_config)

    # Write Level 3 dataset to file if output directory given
    v = pypromice.resources.load_variables(variables)
    m = pypromice.resources.load_metadata(metadata)
    if outpath is not None:
        prepare_and_write(l3, outpath, v, m, '60min')
        prepare_and_write(l3, outpath, v, m, '1D')
        prepare_and_write(l3, outpath, v, m, 'M')
    return l3

def main():
    args = parse_arguments_l2tol3()
    


    _ = get_l2tol3(args.config_folder, 
                   args.inpath, 
                   args.outpath,
                   args.variables, 
                   args.metadata, 
                   args.data_issues_path)
    
if __name__ == "__main__":  
    main()



import logging
from pathlib import Path
from typing import Optional, Dict, Mapping, Sequence

import attrs
import toml


@attrs.define
class StationConfiguration:
    """
    Helper class for storing station specific configurations with respect to

    * Installation specific distance measurements such as height differences between instruments
    * Reference strings such as stid, station_site and wmo_id
    * BUFR export specific parameters

    # TODO: The station related meta data should be fetched from a station specific configuration files in the future or
    # from header data in data source.
    """

    stid: str
    station_site: str = None
    project: Optional[str] = None
    station_type: Optional[str] = None
    wmo_id: Optional[str] = None
    barometer_from_gps: Optional[float] = None
    anemometer_from_sonic_ranger: Optional[float] = None
    temperature_from_sonic_ranger: Optional[float] = None
    height_of_gps_from_station_ground: Optional[float] = None
    sonic_ranger_from_gps: Optional[float] = None
    static_height_of_gps_from_mean_sea_level: Optional[float] = None
    station_relocation: Sequence[str] = attrs.field(factory=list)
    location_type: Optional[str] = None

    # The station data will be exported to BUFR if True. Otherwise, it will only export latest position
    export_bufr: bool = False
    comment: Optional[str] = None

    # skip specific variables for stations
    # If a variable has known bad data, use this collection to skip the variable
    # Note that if a station is not reporting both air temp and pressure it will be skipped,
    # as currently implemented in csv2bufr.min_data_check().
    # ['p_i'], # EXAMPLE
    skipped_variables: Sequence[str] = attrs.field(factory=list)

    positions_update_timestamp_only: bool = False

    @classmethod
    def load_toml(cls, path, skip_unexpected_fields=False):
        config_fields = {field.name for field in attrs.fields(cls)}
        input_dict = toml.load(path)
        unexpected_fields = set(input_dict.keys()) - config_fields
        if unexpected_fields:
            if skip_unexpected_fields:
                logging.info(
                    f"Skipping unexpected fields in toml file {path}: "
                    + ", ".join(unexpected_fields)
                )
                for field in unexpected_fields:
                    input_dict.pop(field)
            else:
                raise ValueError(f"Unexpected fields: {unexpected_fields}")

        return cls(**input_dict)

    def dump_toml(self, path: Path):
        with path.open("w") as fp:
            toml.dump(self.as_dict(), fp)

    def as_dict(self) -> Dict:
        return attrs.asdict(self)


def load_station_configuration_mapping(
    configurations_root_dir: Path,
    **kwargs,
) -> Mapping[str, StationConfiguration]:
    """
    Load station configurations from toml files in configurations_root_dir

    Parameters
    ----------
    configurations_root_dir
        Root directory containing toml files
    kwargs
        Additional arguments to pass to StationConfiguration.load_toml

    Returns
    -------
    Mapping from stid to StationConfiguration

    """
    return {
        config_file.stem: StationConfiguration.load_toml(config_file, **kwargs)
        for config_file in configurations_root_dir.glob("*.toml")
    }


def write_station_configuration_mapping(
    station_configurations: Mapping[str, StationConfiguration],
    configurations_root_dir: Path,
) -> None:
    """
    Write station configurations to toml files in configurations_root_dir

    Parameters
    ----------
    station_configurations
        Mapping from stid to StationConfiguration
    configurations_root_dir
        Output directory

    """
    configurations_root_dir.mkdir(parents=True, exist_ok=True)
    for stid, station_configuration in station_configurations.items():
        with (configurations_root_dir / f"{stid}.toml").open("w") as fp:
            toml.dump(station_configuration.as_dict(), fp)
